import os
import json
import time
from glob import glob
from parquetdb import ParquetDB, config
import pyarrow as pa
from jarvis.db.figshare import data
import pandas as pd

config.logging_config.loggers.parquetdb.level = 'ERROR'
config.apply()

def main():
    dataset_name='alex_pbe_2d_all'
    store_dir = os.path.join(config.data_dir, 'external', 'Jarvis', dataset_name)
    os.makedirs(store_dir, exist_ok=True)
    
    
    # ###########################################################################################
    # # Loading dataset
    # ###########################################################################################
    json_file = glob(os.path.join(store_dir, '*.json'))
    if len(json_file) == 0:
        print("No dataset found, downloading...")
        d = download_jarvis_dataset(dataset_name, store_dir) #choose a name of dataset from above
    
    print("Loading dataset...")
    json_file = glob(os.path.join(store_dir, '*.json'))
    with open(json_file[0], 'r') as f:
        d = json.load(f)
    
    print("converting to pandas dataframe to clean dataset")
    df = pd.DataFrame(d).drop(columns=['id'])


    # ###########################################################################################
    # # Creating database
    # ###########################################################################################
    print('-'*200)
    print("Creating database...")
    print('-'*200)
    db = ParquetDB(dataset_name=dataset_name, dir=store_dir)
    
    if db.dataset_exists():
        db.drop_dataset()
    
    db.create(df)
    
    # Free up memory
    del df
    del d
    # ###########################################################################################
    # # Schema
    # ###########################################################################################
    print('-'*200)
    print("The schema is:\n")
    print('-'*200)
    schema = db.get_schema()
    # Find the longest field name to align the types
    max_name_length = max(len(field.name) for field in schema)
    for field in schema:
        print(f'{field.name:<{max_name_length}} | {field.type}')
    print('-'*200)
    
    del schema
    
    # ###########################################################################################
    # # Full Table read
    # ###########################################################################################
    print('-'*200)
    print("Example of Full Table read")
    print('-'*200)
    time.sleep(0.5)
    print(f"Total allocated bytes: before reading: {pa.total_allocated_bytes() / 10**6} MB")
    table = db.read(rebuild_nested_struct=True)
    
    df=table.to_pandas()
    df.to_csv(os.path.join(store_dir, f'{dataset_name}.csv'), index=False)
    print(df.head())
    
    time.sleep(0.5)
    print(f"Total allocated bytes: after reading: {pa.total_allocated_bytes() / 10**6} MB")
    
    atoms = table['atoms'].combine_chunks()
    print(f"length of the atoms column: {len(atoms)}")
    
    del atoms
    del table
    
    # ###########################################################################################
    # # Batching
    # ###########################################################################################
    print('-'*200)
    print("Example of Batching")
    print('-'*200)
    # Wait for the memory to be freed
    time.sleep(0.5)
    print(f"Total allocated bytes: before creating generator: {pa.total_allocated_bytes() / 10**6} MB")
    generator = db.read(load_format='batches', batch_size=10000, rebuild_nested_struct=True)
    print(f"Total allocated bytes: after creating generator: {pa.total_allocated_bytes() / 10**6} MB")

    for i,record_batch in enumerate(generator):
        # print(record_batch['atoms'].to_pandas())
        batch=record_batch
        print(f"Bytes allocated for batch {i}: {pa.total_allocated_bytes() / 10**6} MB")
        
        del batch


def download_jarvis_dataset(dataset_name: str, store_dir: str):
    os.makedirs(store_dir, exist_ok=True)
    d = data(dataset_name, store_dir=store_dir)
    
    
    # Unzip the downloaded dataset
    zip_files = [f for f in os.listdir(store_dir) if f.endswith('.zip')]
    for zip_file in zip_files:
        import zipfile
        with zipfile.ZipFile(os.path.join(store_dir, zip_file), 'r') as zip_ref:
            zip_ref.extractall(store_dir)
    return d


    
    
if __name__ == "__main__":
    main()



import os
import shutil
import sys
import time
import random
import string
import itertools

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt


import pyarrow as pa
import pyarrow.compute as pc
import pyarrow.dataset as ds
import pyarrow.parquet as pq
import pyarrow.fs as fs

from parquetdb import ParquetDB, config
from parquetdb.utils import general_utils

config.logging_config.loggers.timing.level='ERROR'
config.apply()




if __name__ == '__main__':
    save_dir=os.path.join(config.data_dir, 'benchmarks', 'parquetdb')
    os.makedirs(save_dir, exist_ok=True)

    benchmark_dir=os.path.join(save_dir, 'BenchmarkDB')
    dataset_dir=os.path.join(benchmark_dir, 'parquetdb')

    if os.path.exists(dataset_dir):
        shutil.rmtree(dataset_dir)


    db = ParquetDB(dataset_name='parquetdb', dir=benchmark_dir)
    if db.dataset_exists():
        db.drop_dataset()  # Assuming there's a method to clear the database
    data=general_utils.generate_pydict_data(n_rows=1000000, n_columns=100)
    
    start_time=time.time()
    db.create(data)
    insert_time = time.time() - start_time
    del data
    
    start_time=time.time()
    data = db.read()
    read_time = time.time() - start_time
    del data
    
    db.normalize()


    row_counts = [1, 10, 100, 1000, 10000, 100000, 1000000]
    n_columns=100
    insertion_times = []
    reading_times = []
    update_times=[]

    input_data_generator_dict= {
            'pylist': general_utils.generate_pylist_update_data,
            'pydict': general_utils.generate_pydict_update_data,
            'pandas': general_utils.generate_pandas_update_data,
            'table': general_utils.generate_table_update_data
        }
    
    
    
    results={
        'create_times':[],
        'read_times':[],
        'update_times':[],
        'input_data_type':[],
        'n_rows':[]
    }
    
    
    
    
    for num_rows in row_counts:
        for key, value in input_data_generator_dict.items():
            print(f'Benchmarking {num_rows} rows. Using {key} input data...')
            update_func=value
            
            update_data=update_func(n_rows=num_rows, n_columns=n_columns)
            
            start_time=time.time()
            db.update(update_data)
           
            update_time=time.time()-start_time
            del update_data
            
            print(f'Insertion Time for {num_rows} rows: {insert_time:.4f} seconds')
            print(f'Reading Time for {num_rows} rows: {read_time:.4f} seconds')
            print(f"Updating Time for {num_rows} rows: {update_time:.4f} seconds")
            print('---')
            
            results['create_times'].append(insert_time)
            results['read_times'].append(read_time)
            results['update_times'].append(update_time)
            results['input_data_type'].append(key)
            results['n_rows'].append(num_rows)

    df=pd.DataFrame(results)
    df.to_csv(os.path.join(save_dir, 'parquetdb_input_update_benchmark.csv'), index=False)


import os
import shutil
import sys
import time
import random
import string
import itertools

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1.inset_locator import inset_axes

import pyarrow as pa
import pyarrow.compute as pc
import pyarrow.dataset as ds
import pyarrow.parquet as pq
import pyarrow.fs as fs

from parquetdb import ParquetDB, config
import matplotlib.ticker as ticker

plt.rcParams['axes.labelsize'] = 18
plt.rcParams['axes.titlesize'] = 18
plt.rcParams['xtick.labelsize'] = 14
plt.rcParams['ytick.labelsize'] = 14


db_names=['sqlite','mongodb','parquetdb']
banchmark_dir=os.path.join(config.data_dir, 'benchmarks')
benchmark_dirs=[os.path.join(banchmark_dir, db_name) for db_name in db_names]





def line_style_diff():
    # Initialize a figure and axes
    fig, ax1 = plt.subplots(figsize=(10, 6))


    colors = {
        'create_times': 'blue',
        'read_times': 'red',
    }


    line_styles = {
        'sqlite': 'dotted',
        'mongodb': 'dashdot',
        'parquetdb': 'solid'
    }


    for benchmark_dir in benchmark_dirs:
        basename = os.path.basename(benchmark_dir)
        csv_filename = os.path.join(benchmark_dir, f'{basename}_benchmark.csv')
        
        df = pd.read_csv(csv_filename)

        # Plot create times on the primary y-axis
        ax1.plot(df['n_rows'], df['create_times'], label=f'{basename} create', color=colors['create_times'], linestyle=line_styles[basename])

        # Plot read times on the secondary y-axis
    ax2 = ax1.twinx()

    # ax2.set_xscale('log')
    # ax1.set_yscale('log')
    # ax2.set_yscale('log')
    for benchmark_dir in benchmark_dirs:
        basename = os.path.basename(benchmark_dir)
        csv_filename = os.path.join(benchmark_dir, f'{basename}_benchmark.csv')
        df = pd.read_csv(csv_filename)
        ax2.plot(df['n_rows'], df['read_times'], label=f'{basename} read', color=colors['read_times'], linestyle=line_styles[basename])

    # Set labels and title
    ax1.set_xlabel('Number of Rows')
    ax1.set_ylabel('Create Times (s)', color='blue')
    ax2.set_ylabel('Read Times (s)', color='red')

    # Setting scales
    # ax1.set_xscale('log')
    # ax1.set_yscale('log')
    # ax2.set_yscale('log')


    # Color the spines and tick labels
    ax1.spines['left'].set_color(colors['create_times'])
    ax1.tick_params(axis='y', colors=colors['create_times'])
    ax2.spines['right'].set_color(colors['read_times'])
    ax2.tick_params(axis='y', colors=colors['read_times'])
    
    ax1.grid()

    plt.title('Benchmark Create and Read Times for Different Databases')

    # Combine legends from both axes
    lines_1, labels_1 = ax1.get_legend_handles_labels()
    lines_2, labels_2 = ax2.get_legend_handles_labels()
    ax1.legend(lines_1 + lines_2, labels_1 + labels_2, loc='upper left')

    plt.tight_layout()
    plt.show()
    
    
    
def color_diff_plot():
    # Initialize a figure and axes
    fig, ax1 = plt.subplots(figsize=(10, 6))


    # Define a color map for different databases
    colors = {
        'sqlite': 'blue',
        'mongodb': 'green',
        'parquetdb': 'red'
    }
    
    line_styles = {
        'create_times': 'solid',
        'read_times': 'dashed',
    }

    # Loop through each database benchmark data and plot
    for benchmark_dir in benchmark_dirs:
        basename = os.path.basename(benchmark_dir)
        csv_filename = os.path.join(benchmark_dir, f'{basename}_benchmark.csv')
        
        df = pd.read_csv(csv_filename)

        # Plot create times on the primary y-axis
        ax1.plot(df['n_rows'], df['create_times'], label=f'{basename} create', color=colors[basename], linestyle=line_styles['create_times'])

    # Create twin axis for the secondary y-axis
    ax2 = ax1.twinx()

    for benchmark_dir in benchmark_dirs:
        basename = os.path.basename(benchmark_dir)
        csv_filename = os.path.join(benchmark_dir, f'{basename}_benchmark.csv')
        
        df = pd.read_csv(csv_filename)

        # Plot read times on the secondary y-axis
        ax2.plot(df['n_rows'], df['read_times'], label=f'{basename} read', color=colors[basename], linestyle=line_styles['read_times'])

    # Set labels and title
    ax1.set_xlabel('Number of Rows')
    ax1.set_ylabel('Create Times (s)')
    ax2.set_ylabel('Read Times (s)')
    plt.title('Benchmark Create and Read Times for Different Databases')

    # Ensure both axes have a common scale by linking their limits
    # ax2.set_ylim(ax1.get_ylim())
    # ax1.set_xscale('log')
    # ax1.set_yscale('log')
    # ax2.set_yscale('log')

    # Color the spines and tick labels
    # ax1.spines['left'].set_color('blue')
    # ax1.tick_params(axis='y', colors='blue')
    # ax2.spines['right'].set_color('red')
    # ax2.tick_params(axis='y', colors='red')

    # Set the same linestyle and make the spine thicker for visibility
    ax1.spines['left'].set_linestyle(line_styles['create_times'])
    ax1.spines['left'].set_linewidth(2.5)  # Increase the line width for visibility
    ax2.spines['right'].set_linestyle(line_styles['read_times'])
    ax2.spines['right'].set_linewidth(2.5)  # Increase the line width for visibility

    # Hide the right spine on ax1 and left spine on ax2 to prevent overlap
    ax1.spines['right'].set_visible(False)
    ax2.spines['left'].set_visible(False)

    # Combine legends from both axes
    lines_1, labels_1 = ax1.get_legend_handles_labels()
    lines_2, labels_2 = ax2.get_legend_handles_labels()
    ax1.legend(lines_1 + lines_2, labels_1 + labels_2, loc='upper left')
    
    ax1.grid(True)
    

    plt.tight_layout()
    plt.show()
    

def color_diff_log_two_axes_plot(savefig=None):
    # Initialize a figure and axes
    fig, ax1 = plt.subplots(ncols=2, figsize=(10, 6))


    # Define a color map for different databases
    colors = {
        'sqlite': "#3B9AB2",
        'mongodb': "#EBCC2A",
        'parquetdb': "#F21A00"
    }
    
    line_styles = {
        'create_times': 'solid',
        'read_times': 'dashed',
    }

    # Loop through each database benchmark data and plot
    for benchmark_dir in benchmark_dirs:
        basename = os.path.basename(benchmark_dir)
        csv_filename = os.path.join(benchmark_dir, f'{basename}_benchmark.csv')
        
        df = pd.read_csv(csv_filename)

        # Plot create times on the primary y-axis
        ax1[0].plot(df['n_rows'], df['create_times'], label=f'{basename} create', color=colors[basename], linestyle=line_styles['create_times'])

    # Create twin axis for the secondary y-axis
    ax2 = ax1[0].twinx()

    for benchmark_dir in benchmark_dirs:
        basename = os.path.basename(benchmark_dir)
        csv_filename = os.path.join(benchmark_dir, f'{basename}_benchmark.csv')
        
        df = pd.read_csv(csv_filename)

        # Plot read times on the secondary y-axis
        ax2.plot(df['n_rows'], df['read_times'], label=f'{basename} read', color=colors[basename], linestyle=line_styles['read_times'])

    # Set labels and title
    ax1[0].set_xlabel('Number of Rows')
    ax1[0].set_ylabel('Create Times (s)')
    ax2.set_ylabel('Read Times (s)')
    plt.title('Benchmark Create and Read Times for Different Databases')

    # Ensure both axes have a common scale by linking their limits
    # ax2.set_ylim(ax1[0].get_ylim())
    # ax1[0].set_xscale('log')
    # ax1[0].set_yscale('log')
    # ax2.set_yscale('log')

    # Color the spines and tick labels
    # ax1[0].spines['left'].set_color('blue')
    # ax1[0].tick_params(axis='y', colors='blue')
    # ax2.spines['right'].set_color('red')
    # ax2.tick_params(axis='y', colors='red')

    # Set the same linestyle and make the spine thicker for visibility
    ax1[0].spines['left'].set_linestyle(line_styles['create_times'])
    ax1[0].spines['left'].set_linewidth(2.5)  # Increase the line width for visibility
    ax2.spines['right'].set_linestyle(line_styles['read_times'])
    ax2.spines['right'].set_linewidth(2.5)  # Increase the line width for visibility


    ax1[0].tick_params(axis='both', which='major', length=10, width=2, direction='out')
    ax2.tick_params(axis='both', which='major', length=10, width=2, direction='out')

    # Hide the right spine on ax1[0] and left spine on ax2 to prevent overlap
    ax1[0].spines['right'].set_visible(False)
    ax2.spines['left'].set_visible(False)

    # Combine legends from both axes
    lines_1, labels_1 = ax1[0].get_legend_handles_labels()
    lines_2, labels_2 = ax2.get_legend_handles_labels()
    ax1[0].legend(lines_1 + lines_2, labels_1 + labels_2, loc='upper left')
    
    ax1[0].grid(True)
    
    #######################################################################################################################


    # Loop through each database benchmark data and plot
    for benchmark_dir in benchmark_dirs:
        basename = os.path.basename(benchmark_dir)
        csv_filename = os.path.join(benchmark_dir, f'{basename}_benchmark.csv')
        
        df = pd.read_csv(csv_filename)

        # Plot create times on the primary y-axis
        ax1[1].plot(df['n_rows'], df['create_times'], label=f'{basename} create', color=colors[basename], linestyle=line_styles['create_times'])

    # Create twin axis for the secondary y-axis
    ax2 = ax1[1].twinx()

    for benchmark_dir in benchmark_dirs:
        basename = os.path.basename(benchmark_dir)
        csv_filename = os.path.join(benchmark_dir, f'{basename}_benchmark.csv')
        
        df = pd.read_csv(csv_filename)

        # Plot read times on the secondary y-axis
        ax2.plot(df['n_rows'], df['read_times'], label=f'{basename} read', color=colors[basename], linestyle=line_styles['read_times'])

    # Set labels and title
    ax1[1].set_xlabel('Number of Rows')
    ax1[1].set_ylabel('Create Times (s)')
    ax2.set_ylabel('Read Times (s)')
    plt.title('Benchmark Create and Read Times for Different Databases')

    # Ensure both axes have a common scale by linking their limits
    # ax2.set_ylim(ax1[1].get_ylim())
    ax1[1].set_xscale('log')
    ax1[1].set_yscale('log')
    ax2.set_yscale('log')

    # Color the spines and tick labels
    # ax1[1].spines['left'].set_color('blue')
    # ax1[1].tick_params(axis='y', colors='blue')
    # ax2.spines['right'].set_color('red')
    # ax2.tick_params(axis='y', colors='red')

    # Set the same linestyle and make the spine thicker for visibility
    ax1[1].spines['left'].set_linestyle(line_styles['create_times'])
    ax1[1].spines['left'].set_linewidth(2.5)  # Increase the line width for visibility
    ax2.spines['right'].set_linestyle(line_styles['read_times'])
    ax2.spines['right'].set_linewidth(2.5)  # Increase the line width for visibility

    ax1[1].tick_params(axis='both', which='major', length=10, width=2, direction='out')
    ax2.tick_params(axis='both', which='major', length=10, width=2, direction='out')
    
    # Hide the right spine on ax1[1] and left spine on ax2 to prevent overlap
    ax1[1].spines['right'].set_visible(False)
    ax2.spines['left'].set_visible(False)

    # Combine legends from both axes
    lines_1, labels_1 = ax1[1].get_legend_handles_labels()
    lines_2, labels_2 = ax2.get_legend_handles_labels()
    ax1[1].legend(lines_1 + lines_2, labels_1 + labels_2, loc='upper left')
    
    ax1[1].grid(True)

    plt.tight_layout()
    plt.show()



    
def color_diff_log_inset_plot(savefig=None):
    # Initialize a figure and axes
    fig, ax1 = plt.subplots(figsize=(10, 6))


    # Define a color map for different databases
    # colors = {
    #     'sqlite': "#3B9AB2",
    #     'mongodb': "#EBCC2A",
    #     'parquetdb': "#F21A00"
    # }
    colors = {
        'sqlite': "#FF0000",
        'mongodb': "#F2AD00",
        'parquetdb': "#5BBCD6"
    }
    
    
    line_styles = {
        'create_times': 'solid',
        'read_times': 'dashed',
    }

    # Loop through each database benchmark data and plot
    for benchmark_dir in benchmark_dirs:
        basename = os.path.basename(benchmark_dir)
        csv_filename = os.path.join(benchmark_dir, f'{basename}_benchmark.csv')
        
        df = pd.read_csv(csv_filename)

        # Plot create times on the primary y-axis
        ax1.plot(df['n_rows'], df['create_times'], label=f'{basename} create', color=colors[basename], linestyle=line_styles['create_times'])

    # Create twin axis for the secondary y-axis
    ax2 = ax1.twinx()

    for benchmark_dir in benchmark_dirs:
        basename = os.path.basename(benchmark_dir)
        csv_filename = os.path.join(benchmark_dir, f'{basename}_benchmark.csv')
        
        df = pd.read_csv(csv_filename)

        # Plot read times on the secondary y-axis
        ax2.plot(df['n_rows'], df['read_times'], label=f'{basename} read', color=colors[basename], linestyle=line_styles['read_times'])

    # Set labels and title
    ax1.set_xlabel('Number of Rows')
    ax1.set_ylabel('Create Times (s)')
    ax2.set_ylabel('Read Times (s)')
    
    # Ensure both axes have a common scale by linking their limits
    # ax2.set_ylim(ax1.get_ylim())
    # ax1.set_xscale('log')
    # ax1.set_yscale('log')
    # ax2.set_yscale('log')

    # Color the spines and tick labels
    # ax1.spines['left'].set_color('blue')
    # ax1.tick_params(axis='y', colors='blue')
    # ax2.spines['right'].set_color('red')
    # ax2.tick_params(axis='y', colors='red')

    # Set the same linestyle and make the spine thicker for visibility
    ax1.spines['left'].set_linestyle(line_styles['create_times'])
    ax1.spines['left'].set_linewidth(2.5)  # Increase the line width for visibility
    ax2.spines['right'].set_linestyle(line_styles['read_times'])
    ax2.spines['right'].set_linewidth(2.5)  # Increase the line width for visibility

    # Hide the right spine on ax1 and left spine on ax2 to prevent overlap
    ax1.spines['right'].set_visible(False)
    ax2.spines['left'].set_visible(False)

    ax1.tick_params(axis='both', which='major', length=10, width=2, direction='out')
    # ax1.tick_params(axis='both', which='minor', length=100, width=2, direction='out')
    ax2.tick_params(axis='both', which='major', length=10, width=2, direction='out')
    # ax2.tick_params(axis='both', which='minor', length=10, width=2, direction='out')

    
    ax1.grid(True)
    
    scale=36
    ax_inset = inset_axes(ax1, width=f"{scale}%", height=f"{scale}%", loc="upper left", bbox_to_anchor=(0.05, -0.03,1,1), bbox_transform=ax1.transAxes, borderpad=2)

    # Add an inset with log scale
    # ax_inset = inset_axes(ax1, width="30%", height="30%", loc="upper left", bbox_to_anchor=(0,0,1,1))
    # ax_inset = inset_axes(ax1, width="30%", height="30%", bbox_to_anchor=(0,0,1,1))
    ax_inset.grid(True)
    # Plot the same data with log scales
    for benchmark_dir in benchmark_dirs:
        basename = os.path.basename(benchmark_dir)
        csv_filename = os.path.join(benchmark_dir, f'{basename}_benchmark.csv')
        
        df = pd.read_csv(csv_filename)
        
        ax_inset.plot(df['n_rows'], df['create_times'], label=f'{basename} create', color=colors[basename], linestyle=line_styles['create_times'],  linewidth=2)
        # ax_inset.plot(df['n_rows'], df['read_times'], label=f'{basename} read', color=colors[basename], linestyle='--')

    
    
    # Create a twin axis for the secondary y-axis (inset)
    ax_inset2 = ax_inset.twinx()

    # Plot the read_times in the inset (secondary y-axis)
    for benchmark_dir in benchmark_dirs:
        basename = os.path.basename(benchmark_dir)
        csv_filename = os.path.join(benchmark_dir, f'{basename}_benchmark.csv')
        
        df = pd.read_csv(csv_filename)

        ax_inset2.plot(df['n_rows'], df['read_times'], label=f'{basename} read', color=colors[basename], linestyle=line_styles['read_times'], linewidth=2)

    # Set log scale for both axes in the inset
    ax_inset.set_xscale('log')
    ax_inset.set_yscale('log')
    ax_inset2.set_yscale('log')
    
    # Set labels for inset plot
    ax_inset.set_xlabel('Number of Rows (log)', fontsize=8)
    ax_inset.set_ylabel('Create Time (log)', fontsize=8, labelpad=-2)
    ax_inset2.set_ylabel('Read Time (log)', fontsize=8)
    
    nticks = 9
    maj_loc = ticker.LogLocator(numticks=nticks)
    min_loc = ticker.LogLocator(subs='all', numticks=nticks)
    ax_inset.xaxis.set_major_locator(maj_loc)
    ax_inset.xaxis.set_minor_locator(min_loc)
    
    # Set the same linestyle and make the spine thicker for visibility
    ax_inset.spines['left'].set_linestyle(line_styles['create_times'])
    ax_inset.spines['left'].set_linewidth(2.5)  # Increase the line width for visibility
    ax_inset2.spines['right'].set_linestyle(line_styles['read_times'])
    ax_inset2.spines['right'].set_linewidth(2.5)  # Increase the line width for visibility

    # Hide the right spine on ax1 and left spine on ax2 to prevent overlap
    ax_inset.spines['right'].set_visible(False)
    ax_inset2.spines['left'].set_visible(False)
    
    # ax_inset.tick_params(axis='y', which='major', labelsize=8, length=6, width=1.5, direction='out')
    # ax_inset.tick_params(axis='both', which='minor', direction='out')
    # ax_inset2.tick_params(axis='y', which='major', labelsize=8, length=6, width=1.5, direction='out')
    # ax_inset2.tick_params(axis='both', which='minor', labelsize=8, width=1.5, direction='out')
    ax_inset.tick_params(axis='both', which='major', length=6, width=1.5, direction='out')
    ax_inset.tick_params(axis='x', which='minor', length=3, width=1, direction='out')
    ax_inset.tick_params(axis='y', which='minor', length=3, width=1, direction='out')
    ax_inset2.tick_params(axis='y', which='minor', length=3, width=0, direction='out')

    ax_inset2.tick_params(axis='both', which='major', length=6, width=1.5, direction='out')
    
    # ax_inset.tick_params(axis='x', which='minor', length=6, direction='out')  # Hide minor ticks for the left axis
    # ax_inset2.tick_params(axis='both', which='minor', length=6, direction='out')
    # ax_inset2.tick_params(axis='both', which='minor', length=3, width=1, direction='out')

    # You can also force the ticks to be visible by manually setting locators
    # from matplotlib.ticker import LogLocator
    # ax_inset.yaxis.set_major_locator(LogLocator(base=10.0))
    # ax_inset2.yaxis.set_major_locator(LogLocator(base=10.0))

    # # Optionally, set the minor ticks as well if needed
    # ax_inset.yaxis.set_minor_locator(LogLocator(base=10.0, subs='auto', numticks=10))
    # ax_inset2.yaxis.set_minor_locator(LogLocator(base=10.0, subs='auto', numticks=10))
    
    # ax_inset.set_title('Log Plot', fontsize=10)


    
    
    
    
        
    # Combine legends from both axes
    lines_1, labels_1 = ax1.get_legend_handles_labels()
    lines_2, labels_2 = ax2.get_legend_handles_labels()
    ax1.legend(lines_1 + lines_2, labels_1 + labels_2, loc='upper center',  bbox_to_anchor=(0.15, 0,1,1))
    
    ax1.set_title('Benchmark for Create and Read: \n SQLite, MongoDB, and ParquetDB with 100 integer columns')
    plt.tight_layout()
    
    

    if savefig:
        plt.savefig(savefig)
        return None
    plt.show()
    
    
    

    
    
if __name__ == '__main__':
    from parquetdb.utils import matplotlib_utils
    
    matplotlib_utils.set_palette('Zissou1')
    # color_diff_plot()
    # line_style_diff()
    # color_diff_log_two_axes_plot()
    
    # color_diff_log_inset_plot()
    color_diff_log_inset_plot(savefig=os.path.join(banchmark_dir, 'benchmark_create_read_times.pdf'))
    color_diff_log_inset_plot(savefig=os.path.join(banchmark_dir, 'benchmark_create_read_times.png'))
    


import copy
import os
import logging
import functools
import random
import time

logger = logging.getLogger(__name__)
time_logger = logging.getLogger('timing')

import numpy as np
import pandas as pd
import pyarrow as pa

def timeit(func):
    """
    A decorator that measures and logs the execution time of a function.

    Parameters
    ----------
    func : function
        The function whose execution time is to be measured.

    Returns
    -------
    function
        The wrapped function with timing capabilities.

    Example
    -------
    @timeit
    def example_function():
        # Function logic here
        pass
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        start_time = time.time()
        result = func(*args, **kwargs)
        end_time = time.time()
        elapsed_time = end_time - start_time
        time_logger.debug(f"Function {func.__name__!r} executed in {elapsed_time:.4f} seconds")
        return result

    return wrapper


def is_directory_empty(directory_path:str):
    """
    Checks if a given directory is empty.

    Parameters
    ----------
    directory_path : str
        The path to the directory to check.

    Returns
    -------
    bool
        True if the directory is empty, False otherwise.

    Example
    -------
    >>> is_empty = is_directory_empty('/path/to/directory')
    True
    """

    return not os.listdir(directory_path)


def is_contained(list1, list2):
    """
    Checks if all elements of the first list are contained in the second list.

    Parameters
    ----------
    list1 : list
        The list whose elements are to be checked.
    list2 : list
        The list in which to check for the presence of `list1` elements.

    Returns
    -------
    bool
        True if all elements of `list1` are contained in `list2`, False otherwise.

    Example
    -------
    >>> is_contained([1, 2], [1, 2, 3, 4])
    True
    """
    return set(list1).issubset(set(list2))



def generate_similar_data(template_data, num_entries):
    """
    Generates new data entries based on the structure and values of a template dataset.

    Parameters
    ----------
    template_data : list of dict
        A list of dictionaries containing the template data to base new entries on.
    num_entries : int
        The number of similar data entries to generate.

    Returns
    -------
    list of dict
        A list of generated data entries with variations in the values.

    Example
    -------
    >>> template = [{'id': 0, 'value': 10, 'name': 'item'}]
    >>> new_data = generate_similar_data(template, 5)
    >>> print(new_data)
    [{'id': 0, 'value': 9, 'name': 'item_5'}, {'id': 1, 'value': 11, 'name': 'item_89'}, ...]
    """
    def generate_value(value):
        if isinstance(value, int):
            return random.randint(value - 10, value + 10)
        elif isinstance(value, float):
            return round(random.uniform(value * 0.8, value * 1.2), 2)
        elif isinstance(value, str):
            return f"{value}_{random.randint(1, 100)}"
        elif isinstance(value, dict):
            return {k: generate_value(v) for k, v in value.items()}
        # elif isinstance(value, list):
        #     return [generate_value(item) for item in value]
        elif value is None:
            return None
        else:
            return value

    generated_data = []
    for i in range(num_entries):
        new_entry = copy.deepcopy(random.choice(template_data))
        for key, value in new_entry.items():
            if key == 'id':
                new_entry[key] = i
            else:
                new_entry[key] = generate_value(value)
        generated_data.append(new_entry)

    return generated_data



def generate_pydict_data(n_rows=100, n_columns=100, min_value=0, max_value=100000):
    data = {}
    for i in range(n_columns):
        column_name = f"column_{i}"
        data[column_name] = [random.randint(min_value, max_value) for _ in range(n_rows)]

    return data

def generate_pydict_update_data(n_rows=100, n_columns=100, min_value=0, max_value=100):
    data = {}
    for i in range(n_columns):
        column_name = f"column_{i}"
        data[column_name] = [random.randint(min_value, max_value) for _ in range(n_rows)]
    data['id'] = [i for i in range(n_rows)]
    return data

def generate_pylist_data(n_rows=100, n_columns=100):
    data=[]
    for _ in range(n_rows):
        data.append({f'column_{i}':random.randint(0, 100000) for i in range(n_columns)})
    return data

def generate_pylist_update_data(n_rows=100, n_columns=100):
    data = []
    for i in range(n_rows):
        row = {'id': i}  # Unique identifier for each row
        row.update({f'column_{j}': random.randint(0, 100000) for j in range(n_columns)})
        data.append(row)
    return data

def generate_pandas_data(n_rows=100, n_columns=100):
    df=pd.DataFrame(generate_pydict_data(n_rows=n_rows, n_columns=n_columns))
    return df
    
def generate_pandas_update_data(n_rows=100, n_columns=100):
    df=pd.DataFrame(generate_pydict_update_data(n_rows=n_rows, n_columns=n_columns))
    return df

def generate_table_data(n_rows=100, n_columns=100):
    data=generate_pydict_data(n_rows=n_rows, n_columns=n_columns)
    table=pa.Table.from_pydict(data)
    return table

def generate_table_update_data(n_rows=100, n_columns=100):
    data=generate_pydict_update_data(n_rows=n_rows, n_columns=n_columns)
    table=pa.Table.from_pydict(data)
    return table



import unittest
import numpy as np
import pyarrow as pa
import pandas as pd
from dataclasses import dataclass

from parquetdb.core.types import (
    PythonObjectArrowType,
    PythonObjectPandasArray,
    PythonObjectPandasDtype
)

@dataclass
class DummyClass:
    x: int
    y: str
    
    def __eq__(self, other):
        if not isinstance(other, DummyClass):
            return False
        return self.x == other.x and self.y == other.y
    
    def __call__(self, x):
        return self.x * x

def dummy_function(x):
    return x ** 2

class TestPythonObjectTypes(unittest.TestCase):
    
    def setUp(self):
        # Create test objects of different types
        self.dummy_obj = DummyClass(x=1, y="test")
        self.dummy_func = dummy_function
        self.lambda_func = lambda x: x + 1
        
        # Create test data with different types of objects
        self.test_objects = [
            self.dummy_obj,
            self.dummy_func,
            self.lambda_func,
            None,
            [1, 2, 3],
            {"a": 1, "b": 2}
        ]
        
        self.test_data={'dummy_obj':[self.dummy_obj for _ in range(5)], 
                        'dummy_func':[self.dummy_func for _ in range(5)], 
                        'lambda_func':[self.lambda_func for _ in range(5)], 
                        'none':[None for _ in range(5)], 
                        'list':[[1, 2, 3] for _ in range(5)], 
                        'dict':[{"a": 1, "b": 2} for _ in range(5)]}
        
        self.test_data_custom={'dummy_obj':PythonObjectPandasArray([self.dummy_obj for _ in range(5)]), 
                        'dummy_func':PythonObjectPandasArray([self.dummy_func for _ in range(5)]), 
                        'lambda_func':PythonObjectPandasArray([self.lambda_func for _ in range(5)]), 
                        'none':[None for _ in range(5)], 
                        'list':[[1, 2, 3] for _ in range(5)], 
                        'dict':[{"a": 1, "b": 2} for _ in range(5)]}

    def test_pandas_array_creation(self):
        # Test creating PythonObjectPandasArray
        
        for key in self.test_data:
            arr = PythonObjectPandasArray(self.test_data[key])
            
            # Verify length and contents
            self.assertEqual(len(arr), len(self.test_data[key]))
            
            self.assertEqual(arr.shape, (len(self.test_data[key]),))        
            # Test that functions still work after storage
            
            if key=='dummy_obj':
                self.assertEqual(arr[0](4), 4)
                self.assertEqual(arr[0].x, 1)
                self.assertEqual(arr[0].y, 'test')
            elif key=='dummy_func':
                self.assertEqual(arr[0](4), 16)
            elif key=='lambda_func':
                self.assertEqual(arr[0](4), 5)
        

    def test_arrow_conversion(self):
        # Test conversion from PythonObjectPandasArray to Arrow Array
       for key in self.test_data:
            arr = PythonObjectPandasArray(self.test_data[key])
            arrow_arr = pa.array(arr, type=PythonObjectArrowType())
            
            self.assertEqual(arrow_arr.type, PythonObjectArrowType())
        
            series=arrow_arr.to_pandas(split_blocks=True, self_destruct=True)
            
            if key=='dummy_obj':
                self.assertEqual(series[0].x, 1)
                self.assertEqual(series[0].y, 'test')
                self.assertEqual(series[2].__class__, type(self.dummy_obj))
            elif key=='dummy_func':
                self.assertEqual(series[0](4), 16)
                self.assertEqual(series[2].__class__, self.test_data[key][2].__class__)
            elif key=='lambda_func':
                self.assertEqual(series[0](4), 5)
                self.assertEqual(series[2].__class__, self.test_data[key][2].__class__)
            elif key=='none':
                self.assertEqual(series[0], None)
                self.assertEqual(series[2], None)
            elif key=='list':
                self.assertEqual(series[0], [1, 2, 3])
                self.assertEqual(series[2], [1, 2, 3])
            elif key=='dict':
                self.assertEqual(series[0], {"a": 1, "b": 2})
                self.assertEqual(series[2], {"a": 1, "b": 2})
    


    def test_dataframe_integration(self):
        # Test integration with pandas DataFrame
        df_initial = pd.DataFrame(self.test_data_custom)
        
        self.assertEqual(df_initial['dummy_obj'].dtype, PythonObjectPandasDtype())
        self.assertEqual(df_initial['dummy_func'].dtype, PythonObjectPandasDtype())
        self.assertEqual(df_initial['lambda_func'].dtype, PythonObjectPandasDtype())
        self.assertEqual(df_initial['none'].dtype, object)
        self.assertEqual(df_initial['list'].dtype, object)
        self.assertEqual(df_initial['dict'].dtype, object)
        
        table=pa.Table.from_pandas(df_initial)
        
        self.assertEqual(table['dummy_obj'].type, PythonObjectArrowType())
        self.assertEqual(table['dummy_func'].type, PythonObjectArrowType())
        self.assertEqual(table['lambda_func'].type, PythonObjectArrowType())
        self.assertEqual(table['none'].type, pa.null())
        self.assertEqual(table['list'].type, pa.list_(pa.int64()))
        self.assertEqual(table['dict'].type, pa.struct([pa.field('a', pa.int64()), pa.field('b', pa.int64())]))
        
        
        df=table.to_pandas(split_blocks=True, self_destruct=True)
        
        self.assertEqual(df['dummy_obj'].dtype, PythonObjectPandasDtype())
        self.assertEqual(df['dummy_func'].dtype, PythonObjectPandasDtype())
        self.assertEqual(df['lambda_func'].dtype, PythonObjectPandasDtype())
        self.assertEqual(df['none'].dtype, object)
        self.assertEqual(df['list'].dtype, object)
        self.assertEqual(df['dict'].dtype, object)
        
        for irow, row in df.iterrows():
            self.assertEqual(row['dummy_obj'], df_initial['dummy_obj'][irow])


    def test_array_operations(self):
        arr = PythonObjectPandasArray(self.test_objects)
        
        # Test slicing
        slice_arr = arr[1:4]
        self.assertEqual(len(slice_arr), 3)
        self.assertEqual(slice_arr[0](2), 4)  # dummy_function(2)
        
        # Test isna
        na_mask = arr.isna()
        self.assertTrue(na_mask[3])  # None should be identified as NA
        self.assertFalse(na_mask[0])  # DummyClass instance should not be NA
        
        # Test iteration
        for i, obj in enumerate(arr):
            self.assertEqual(obj, self.test_objects[i])

    def test_concatenation(self):
        arr1 = PythonObjectPandasArray(self.test_objects[:3])
        arr2 = PythonObjectPandasArray(self.test_objects[3:])
        
        # Test concatenation
        concatenated = PythonObjectPandasArray._concat_same_type([arr1, arr2])
        
        
        df_1=pd.DataFrame(self.test_data_custom)
        df_2=pd.DataFrame(self.test_data_custom)
        df_concat=pd.concat([df_1, df_2])
        
        self.assertEqual(df_concat.shape, (10, 6))
        # # Verify length and contents
        # self.assertEqual(len(concatenated), len(self.test_objects))
        # for i, obj in enumerate(self.test_objects):
        #     if obj is None:
        #         self.assertIsNone(concatenated[i])
        #     else:
        #         self.assertEqual(concatenated[i], obj)

    def test_copy(self):
        arr = self.test_data_custom['dummy_obj']
        arr_copy = arr.copy()
        
        # Verify it's a true copy
        self.assertEqual(len(arr), len(arr_copy))
        for i in range(len(arr)):
            self.assertEqual(arr[i], arr_copy[i])
        
        # Verify that modifying the copy doesn't affect the original
        arr_copy._data[0] = None
        self.assertNotEqual(arr[0], arr_copy[0])

if __name__ == '__main__':
    
    # unittest.TextTestRunner().run(TestPythonObjectTypes('test_pandas_array_creation'))
    # unittest.TextTestRunner().run(TestPythonObjectTypes('test_arrow_conversion'))
    # unittest.TextTestRunner().run(TestPythonObjectTypes('test_dataframe_integration'))
    # unittest.TextTestRunner().run(TestPythonObjectTypes('test_array_operations'))
    # unittest.TextTestRunner().run(TestPythonObjectTypes('test_concatenation'))
    # unittest.TextTestRunner().run(TestPythonObjectTypes('test_copy'))
    unittest.main()



import json
import logging
import os
import shutil
import tempfile
import time
import unittest

import numpy as np
import pandas as pd
import pyarrow as pa
import pyarrow.compute as pc
import pyarrow.parquet as pq

from parquetdb import ParquetDB, config
from parquetdb.core import types
from parquetdb.core.parquetdb import LoadConfig, NormalizeConfig
from parquetdb.utils import pyarrow_utils

logger = logging.getLogger("tests")

# config.logging_config.loggers.timing.level='ERROR'
config.logging_config.loggers.parquetdb.level = "DEBUG"
# config.logging_config.loggers.tests.level='ERROR'
# config.apply()

with open(os.path.join(config.tests_dir, "data", "alexandria_test.json"), "r") as f:
    alexandria_data = json.load(f)


class TestParquetDB(unittest.TestCase):
    def setUp(self):
        # Create a temporary directory for the database
        self.temp_dir = tempfile.mkdtemp()

        self.db = ParquetDB(db_path=os.path.join(self.temp_dir, "test_db"))

        # Create some test data
        self.test_data = [
            {"name": "Alice", "age": 30},
            {"name": "Bob", "age": 25},
            {"name": "Charlie", "age": 35},
        ]
        logger.debug(f"Test data: {self.test_data}")
        self.test_df = pd.DataFrame(self.test_data)

    def tearDown(self):
        # Remove the temporary directory after the test
        try:
            shutil.rmtree(self.temp_dir)
        except Exception as e:
            logger.error(f"Error during cleanup: {e}")
        # For some reason, there are race conditions when
        # deleting the directory and performaing another test
        # time.sleep(0.1)

    def test_create_and_read(self):
        logger.info("Testing create and read")
        # Test creating data and reading it back
        data = [{"name": "Alice", "age": 30}, {"name": "Bob", "age": 25}]
        self.db.create(data)
        table = self.db.read()
        df = table.to_pandas()
        logger.debug(f"DataFrame:\n{df}")
        data = [{"name": "Alice", "age": 30}, {"name": "Bob", "age": 25}]
        self.db.create(data)

        # Read the data back
        table = self.db.read()
        df = table.to_pandas()
        logger.debug(f"DataFrame:\n{df}")

        # Assertions
        self.assertEqual(len(df), 4)
        self.assertIn("name", df.columns)
        self.assertIn("age", df.columns)
        self.assertEqual(df[df["age"] == 30].iloc[0]["name"], "Alice")
        self.assertEqual(df[df["age"] == 25].iloc[0]["name"], "Bob")

        logger.info("Test create and read passed")

    def test_update(self):
        logger.info("Testing update")
        # Test updating existing records

        data = [{"name": "Charlie", "age": 28}, {"name": "Diana", "age": 32}]
        self.db.create(data)

        # Read the data back
        result = self.db.read()
        df = result.to_pandas()
        logger.debug(f"DataFrame:\n{df}")

        # Update the age of 'Charlie'
        update_data = [{"id": 0, "age": 29}]
        self.db.update(update_data)

        # Read the data back
        result = self.db.read()
        df = result.to_pandas()
        logger.debug(f"DataFrame:\n{df}")

        # Assertions
        self.assertEqual(df[df["name"] == "Charlie"].iloc[0]["age"], 29)
        self.assertEqual(df[df["name"] == "Diana"].iloc[0]["age"], 32)
        logger.info("Test update passed")

    def test_delete(self):
        # Test deleting records
        data = [{"name": "Eve", "age": 35}, {"name": "Frank", "age": 40}]
        self.db.create(data)

        # Delete 'Eve'
        self.db.delete(
            ids=[0],
        )

        # Read the data back
        result = self.db.read()
        df = result.to_pandas()

        # Assertions
        self.assertEqual(len(df), 1)
        self.assertEqual(df.iloc[0]["name"], "Frank")

    def test_create_and_normalize(self):
        # Step 1: Create data without normalization
        self.db.create(data=self.test_df, normalize_dataset=False)

        # Step 2: Verify that data has been written to the dataset directory
        dataset_files = self.db.get_current_files()
        self.assertGreater(
            len(dataset_files),
            0,
            "No parquet files found after create without normalization.",
        )

        # Load the data to check its state before normalization
        loaded_data = self.db.read()
        self.assertEqual(
            loaded_data.num_rows,
            len(self.test_data),
            "Mismatch in row count before normalization.",
        )

        # Step 3: Run normalization. Will normalize to 1 row per file and 1 row per group
        self.db.normalize(
            normalize_config=NormalizeConfig(
                max_rows_per_file=1, min_rows_per_group=1, max_rows_per_group=1
            )
        )

        # Step 4: Verify that the data has been normalized (e.g., consistent row distribution)
        normalized_data = self.db.read()
        self.assertEqual(
            normalized_data.num_rows, 3, "Mismatch in row count after normalization."
        )

        # Additional checks to ensure normalization affects file structure (if applicable)
        normalized_files = self.db.get_current_files()
        self.assertGreaterEqual(
            len(normalized_files), 3, "No files found after normalization."
        )

    def test_filters(self):
        # Test reading data with filters
        data = [
            {"name": "Grace", "age": 22},
            {"name": "Heidi", "age": 27},
            {"name": "Ivan", "age": 35},
        ]
        self.db.create(data)

        # Apply filter to get people older than 25
        age_filter = pc.field("age") > 25
        result = self.db.read(filters=[age_filter])
        df = result.to_pandas()

        # Assertions
        self.assertEqual(len(df), 2)
        self.assertListEqual(df["name"].tolist(), ["Heidi", "Ivan"])

    def test_add_new_field(self):
        # Test adding data with a new field and ensure schema evolves
        # This also test that if new incoming data doesn't have a
        # field that is already in the schema, a null value of the correct type is added
        data = [{"name": "Judy", "age": 29}]
        self.db.create(data)

        # Add new data with an additional field
        new_data = [{"name": "Karl", "occupation": "Engineer"}]
        self.db.create(new_data)

        # Read back the data
        table = self.db.read()
        df = table.to_pandas()

        # Assertions
        self.assertIn("occupation", df.columns)
        self.assertEqual(df[df["name"] == "Karl"].iloc[0]["occupation"], "Engineer")
        self.assertTrue(pd.isnull(df[df["name"] == "Judy"].iloc[0]["occupation"]))
        self.assertTrue(np.isnan(df[df["name"] == "Karl"].iloc[0]["age"]))

    def test_nested_data_handling(self):

        py_lattice_1 = [
            [[1.1, 0, 0], [0, 1, 0], [0, 0, 1]],
            [[1, 0, 0], [0, 1, 0], [0, 0, 1]],
        ]
        py_lattice_2 = [
            [[2.1, 0, 0], [0, 2, 0], [0, 0, 2]],
            [[2, 0, 0], [0, 2, 0], [0, 0, 2]],
        ]

        current_data = [
            {
                "b": {"x": 10, "y": 20},
                "c": {},
                "e": 5,
                "lattice": py_lattice_1,
                "unordered_list_string": ["hi", "reree"],
                "unordered_list_int": [0, 1, 3],
            },
            {
                "b": {"x": 30, "y": 40},
                "c": {},
                "lattice": py_lattice_2,
                "unordered_list_string": ["reree"],
                "unordered_list_int": [0, 1],
            },
            {"b": {"x": 30, "y": 40}, "c": {}},
            {"b": {"x": 30, "y": 40}, "c": {}},
            {"b": {"x": 30, "y": 40}, "c": {}},
            {
                "b": {"x": 30, "y": 40},
                "c": {},
                "lattice": py_lattice_2,
                "unordered_list_string": ["reree"],
                "unordered_list_int": [0, 1],
            },
        ]

        incoming_data = [
            {"id": 2, "b": {"x": 10, "y": 20}, "c": {}},
            {
                "id": 0,
                "b": {"y": 50, "z": {"a": 1.1, "b": 3.3}},
                "c": {},
                "d": 1,
                "e": 6,
                "lattice": py_lattice_2,
                "unordered_list_string": ["updated"],
            },
            {
                "id": 3,
                "b": {"y": 50, "z": {"a": 4.5, "b": 5.3}},
                "c": {},
                "d": 1,
                "e": 6,
                "lattice": py_lattice_2,
                "unordered_list_string": ["updated"],
            },
        ]

        self.db.create(current_data)

        # Initial read before the update
        table = self.db.read()
        df = table.to_pandas()

        # logger.debug(f"DataFrame:\n{df.head()}")
        # Assert the shape before the update (confirming initial data structure)
        assert df.shape[0] == len(
            current_data
        ), f"Expected {len(current_data)} rows before update, but got {df.shape[0]}"
        assert df.loc[0, "b.x"] == 10
        assert df.loc[0, "b.y"] == 20
        assert np.isnan(df.loc[0, "c.dummy_field"])
        assert df.loc[0, "e"] == 5
        assert np.array_equal(
            table["lattice"].combine_chunks().to_numpy_ndarray()[0],
            np.array(py_lattice_1),
        ), "Row 0 lattice data mismatch after update"
        assert (
            df.loc[0, "unordered_list_string"][0] == "hi"
        ), "Row 0 'unordered_list_string' did not update correctly"
        assert (
            df.loc[0, "unordered_list_string"][1] == "reree"
        ), "Row 0 'unordered_list_string' did not update correctly"
        assert (
            df.loc[0, "unordered_list_int"][0] == 0
        ), "Row 0 'unordered_list_int' did not update correctly"
        assert (
            df.loc[0, "unordered_list_int"][1] == 1
        ), "Row 0 'unordered_list_int' did not update correctly"
        assert (
            df.loc[0, "unordered_list_int"][2] == 3
        ), "Row 0 'unordered_list_int' did not update correctly"

        assert df.loc[1, "b.x"] == 30
        assert df.loc[1, "b.y"] == 40
        assert np.isnan(df.loc[1, "c.dummy_field"])
        assert np.isnan(df.loc[1, "e"])
        assert np.array_equal(
            table["lattice"].combine_chunks().to_numpy_ndarray()[1],
            np.array(py_lattice_2),
        ), "Row 0 lattice data mismatch after update"
        assert (
            df.loc[1, "unordered_list_string"][0] == "reree"
        ), "Row 0 'unordered_list_string' did not update correctly"
        assert (
            df.loc[1, "unordered_list_int"][0] == 0
        ), "Row 0 'unordered_list_int' did not update correctly"
        assert (
            df.loc[1, "unordered_list_int"][1] == 1
        ), "Row 0 'unordered_list_int' did not update correctly"

        assert df.loc[2, "b.x"] == 30
        assert df.loc[2, "b.y"] == 40
        assert np.isnan(df.loc[2, "c.dummy_field"])
        assert np.isnan(df.loc[2, "e"])
        assert np.isnan(df.loc[2, "e"])
        assert np.isnan(df.loc[2, "e"])
        assert np.isnan(df.loc[2, "lattice"]).all()
        assert df.loc[2, "unordered_list_string"] == None
        assert df.loc[2, "unordered_list_int"] == None

        assert df.shape[0] == len(
            current_data
        ), f"Expected {len(current_data)} rows before update, but got {df.shape[0]}"
        assert df.loc[5, "b.x"] == 30
        assert df.loc[5, "b.y"] == 40
        assert np.isnan(df.loc[5, "c.dummy_field"])
        assert np.isnan(df.loc[5, "e"])
        assert np.array_equal(
            table["lattice"].combine_chunks().to_numpy_ndarray()[5],
            np.array(py_lattice_2),
        ), "Row 0 lattice data mismatch after update"
        assert (
            df.loc[5, "unordered_list_string"][0] == "reree"
        ), "Row 0 'unordered_list_string' did not update correctly"
        assert (
            df.loc[5, "unordered_list_int"][0] == 0
        ), "Row 0 'unordered_list_int' did not update correctly"
        assert (
            df.loc[5, "unordered_list_int"][1] == 1
        ), "Row 0 'unordered_list_int' did not update correctly"

        # Perform the update
        self.db.update(incoming_data)

        # Read back the data after the update
        table = self.db.read()
        df = table.to_pandas()

        # Assert that the number of rows remains the same (no new rows added)
        assert df.shape[0] == len(
            current_data
        ), f"Expected {len(current_data)} rows after update, but got {df.shape[0]}"

        # Check if the updates for `id:0` have been applied correctly
        assert df.loc[0, "b.y"] == 50
        assert df.loc[0, "b.x"] == 10
        assert df.loc[0, "b.z.a"] == 1.1
        assert df.loc[0, "b.z.b"] == 3.3
        assert df.loc[0, "e"] == 6, "Row 0 column 'e' values mismatch after update"
        assert df.loc[0, "unordered_list_string"] == [
            "updated"
        ], "Row 0 'unordered_list_string' did not update correctly"

        assert np.array_equal(
            table["lattice"].combine_chunks().to_numpy_ndarray()[0],
            np.array(py_lattice_2),
        ), "Row 0 lattice data mismatch after update"

        # Check if the updates for `id:3` have been applied correctly
        assert df.loc[3, "b.y"] == 50
        assert df.loc[3, "b.x"] == 30
        assert df.loc[3, "b.z.a"] == 4.5
        assert df.loc[3, "b.z.b"] == 5.3
        assert df.loc[3, "e"] == 6, "Row 3 column 'e' values mismatch after update"
        assert df.loc[3, "unordered_list_string"] == [
            "updated"
        ], "Row 3 'unordered_list_string' did not update correctly"
        assert np.array_equal(
            table["lattice"].combine_chunks().to_numpy_ndarray()[3],
            np.array(py_lattice_2),
        ), "Row 3 lattice data mismatch after update"

        # Check that rows without 'id' in incoming data remain unchanged (e.g., row 1)
        assert df.loc[1, "b.y"] == 40
        assert df.loc[1, "b.x"] == 30
        assert df.loc[1, "unordered_list_string"] == [
            "reree"
        ], "Row 1 'unordered_list_string' unexpectedly changed"
        assert np.array_equal(
            table["lattice"].combine_chunks().to_numpy_ndarray()[1],
            np.array(py_lattice_2),
        ), "Row 1 lattice data unexpectedly changed"

    def test_get_schema(self):
        # Test retrieving the schema
        data = [{"name": "Liam", "age": 45}]
        self.db.create(data)
        schema = self.db.get_schema()

        # Assertions
        self.assertIn("name", schema.names)
        self.assertIn("age", schema.names)
        self.assertIn("id", schema.names)

    def test_read_specific_columns(self):
        # Test reading specific columns
        data = [
            {"name": "Mia", "age": 30, "city": "New York"},
            {"name": "Noah", "age": 35, "city": "San Francisco"},
        ]
        self.db.create(data)

        # Read only the 'name' column
        result = self.db.read(columns=["name"])
        df = result.to_pandas()

        # Assertions
        self.assertEqual(len(df.columns), 1)
        self.assertIn("name", df.columns)
        self.assertNotIn("age", df.columns)
        self.assertNotIn("city", df.columns)

    def test_batch_reading(self):
        # Test reading data in batches
        data = [{"name": f"Person {i}", "age": i} for i in range(100)]
        self.db.create(data)

        # Read data in batches of 20
        batches = self.db.read(batch_size=20, load_format="batches")

        # Assertions
        batch_count = 0
        total_rows = 0
        for batch in batches:
            batch_count += 1
            total_rows += batch.num_rows
            self.assertLessEqual(batch.num_rows, 20)
        self.assertEqual(batch_count, 5)
        self.assertEqual(total_rows, 100)

    def test_update_schema(self):
        # Test updating the schema of the table
        data = [{"name": "Olivia", "age": 29}]
        self.db.create(data)

        # Read back the data
        result = self.db.read()
        df = result.to_pandas()

        logger.debug(f"DataFrame:\n{df}")

        # Update the 'age' field to be a float instead of int
        new_field = pa.field("age", pa.float64())
        field_dict = {"age": new_field}
        self.db.update_schema(field_dict=field_dict)

        # Read back the data
        result = self.db.read()
        df = result.to_pandas()

        logger.debug(f"DataFrame:\n{df}")

        # Assertions
        self.assertEqual(df["age"].dtype, "float64")

    def test_update_with_new_field_included(self):
        # Test updating the schema of the table
        data = [
            {"name": "Mia", "age": 30, "city": "New York"},
            {"name": "Noah", "age": 35, "city": "San Francisco"},
        ]
        self.db.create(data)

        # Update the 'Mia' record to include a new field and change age to 60
        data = [{"id": 0, "age": 60, "state": "NY"}]
        self.db.update(data)

        # Read back the data
        result = self.db.read()
        df = result.to_pandas()

        # Assertions
        self.assertEqual(df.iloc[0]["state"], "NY")
        self.assertEqual(df.iloc[1]["state"], None)
        self.assertEqual(df.iloc[0]["age"], 60)
        self.assertEqual(df.iloc[1]["age"], 35)

    def test_delete_nonexistent_id(self):
        # Test deleting an ID that doesn't exist
        data = [{"name": "Peter", "age": 50}]
        self.db.create(data)

        # Attempt to delete a non-existent ID
        self.db.delete(ids=[999])

        # Read back the data
        result = self.db.read()
        df = result.to_pandas()

        # Assertions
        self.assertEqual(len(df), 1)
        self.assertEqual(df.iloc[0]["name"], "Peter")

    def test_metadata(self):
        metadata = self.db.get_metadata()
        self.db.set_metadata({"class": "test"})
        assert self.db.get_metadata()["class"] == "test"

        self.db.create(data=self.test_data)

        metadata = self.db.get_metadata()
        assert metadata == {"class": "test"}

        self.db.create(
            data=self.test_data, metadata={"key1": "value1", "key2": "value2"}
        )
        # Should return metadata dictionary (can be empty)
        metadata = self.db.get_metadata()

        self.assertEqual(metadata["key1"], "value1")
        self.assertEqual(metadata["key2"], "value2")

        self.db.set_metadata({"key3": "value3", "key4": "value4"})
        metadata = self.db.get_metadata()

        self.assertEqual(metadata["class"], "test")
        self.assertEqual(metadata["key1"], "value1")
        self.assertEqual(metadata["key2"], "value2")
        self.assertEqual(metadata["key3"], "value3")
        self.assertEqual(metadata["key4"], "value4")

        # Testing set_metadata with update=False
        self.db.set_metadata({"key5": "value5", "key6": "value6"}, update=False)
        assert self.db.get_metadata() == {"key5": "value5", "key6": "value6"}

        # Testing set_field_metadata with update=True

        fields_metadata = {"name": {"key1": "value1", "key2": "value2"}}
        self.db.set_field_metadata(fields_metadata=fields_metadata)
        fields_metadata = self.db.get_field_metadata(field_names=["name"])
        assert fields_metadata["name"] == {"key1": "value1", "key2": "value2"}

        # Testing set_field_metadata with update=True with existing metadata
        fields_metadata = {"name": {"key3": "value3", "key4": "value4"}}
        self.db.set_field_metadata(fields_metadata=fields_metadata, update=True)
        fields_metadata = self.db.get_field_metadata(
            field_names=["name"], return_bytes=True
        )
        assert fields_metadata["name"] == {
            b"key1": b"value1",
            b"key2": b"value2",
            b"key3": b"value3",
            b"key4": b"value4",
        }

        # Testing set_field_metadata with update=False
        fields_metadata = {"name": {"key3": "value3", "key4": "value4"}}
        self.db.set_field_metadata(fields_metadata=fields_metadata, update=False)
        fields_metadata = self.db.get_field_metadata(
            field_names="name", return_bytes=True
        )
        assert fields_metadata["name"] == {b"key3": b"value3", b"key4": b"value4"}

    def test_drop_dataset(self):
        self.db.create(data=self.test_data)
        # Drop the table and check if it no longer exists
        self.db.drop_dataset()

    def test_rename_dataset(self):
        self.db.create(data=self.test_data)
        # Rename the table and check if the new name exists
        self.db.rename_dataset("renamed_table", remove_dest=True)

        assert self.db.dataset_name == "renamed_table"

    def test_export_dataset(self):
        self.db.create(data=self.test_data)
        # Export the table to CSV
        export_path = os.path.join(self.temp_dir, "exported_table.csv")
        self.db.export_dataset(export_path, format="csv")
        self.assertTrue(os.path.exists(export_path))

        # Verify the exported data
        exported_df = pd.read_csv(export_path)
        original_df = self.db.read().to_pandas()
        pd.testing.assert_frame_equal(original_df, exported_df)

        # Export to an unsupported format
        with self.assertRaises(ValueError):
            self.db.export_dataset(export_path, format="xlsx")

    def test_rename_fields(self):
        self.db.create(data=self.test_data)
        self.db.rename_fields({"name": "first_name"})
        schema = self.db.get_schema()
        assert "first_name" in schema.names

    def test_sort_fields(self):
        self.db.create(data=self.test_data)
        self.db.sort_fields()
        schema = self.db.get_schema()
        schema_names = schema.names

        assert schema_names[0] == "age"
        assert schema_names[1] == "id"
        assert schema_names[2] == "name"

    def test_fixed_shape_tensor(self):
        a = np.eye(3).tolist()
        data_1 = [{"a": 3}, {"a": 4}]
        data_2 = [{"2d_array": a}]
        update_data = [{"2d_array": a, "id": 0}]
        self.db.create(data_1)
        self.db.create(data_2)

        table = self.db.read()
        assert table["2d_array"].combine_chunks().to_numpy_ndarray().shape == (3, 3, 3)

        self.db.update(update_data)
        table = self.db.read()
        arrays = table["2d_array"].combine_chunks().to_numpy_ndarray()
        ids = table["id"].combine_chunks().to_pylist()

        assert ids == [0, 1, 2]

        assert np.array_equal(arrays[0], np.eye(3))
        assert np.array_equal(arrays[2], np.eye(3))

    def test_rebuild_nested(self):
        data = [{"a": 1, "b": {"c": 2, "d": 3}}, {"a": 4, "b": {"c": 5, "d": 6}}]
        self.db.create(data)
        table = self.db.read(rebuild_nested_struct=True)

        assert pa.types.is_struct(table["b"].type)
        assert table["b"].type.num_fields == 2

        assert table["b"].combine_chunks().to_pylist()[0]["c"] == 2
        assert table["b"].combine_chunks().to_pylist()[0]["d"] == 3

    def test_update_maintains_existing_extension_arrays(self):
        data_1 = [{"pbc": [1, 0, 0]}, {"pbc": [0, 1, 0]}]
        self.db.create(data_1)

        table = self.db.read()

        assert table["pbc"].combine_chunks().to_numpy_ndarray().tolist() == [
            [1, 0, 0],
            [0, 1, 0],
        ]
        data_2 = [{"id": 0, "density": 5}]

        self.db.update(data_2)
        table = self.db.read()
        assert table["pbc"].combine_chunks().to_numpy_ndarray().tolist() == [
            [1, 0, 0],
            [0, 1, 0],
        ]

    def test_update_maintains_existing_extension_arrays_batches(self):
        data_1 = [{"pbc": [1, 0, 0]}, {"pbc": [0, 1, 0]}]
        self.db.create(data_1)

        table = self.db.read()

        assert table["pbc"].combine_chunks().to_numpy_ndarray().tolist() == [
            [1, 0, 0],
            [0, 1, 0],
        ]
        data_2 = [{"id": 0, "density": 5}]

        self.db.update(
            data_2,
            normalize_config=NormalizeConfig(load_format="batches", batch_size=1),
        )
        table = self.db.read()
        assert table["pbc"].combine_chunks().to_numpy_ndarray().tolist() == [
            [1, 0, 0],
            [0, 1, 0],
        ]

    def test_update_on_key(self):
        data_1 = [
            {"material_id": 1, "material_name": "material_1"},
            {"material_id": 2, "material_name": "material_2"},
        ]
        self.db.create(data_1)

        table = self.db.read()

        assert table["material_id"].combine_chunks().to_pylist() == [1, 2]
        data_2 = [{"material_id": 1, "material_name": "material_1_updated"}]

        self.db.update(data_2, update_keys="material_id")
        table = self.db.read()
        assert table["material_id"].combine_chunks().to_pylist() == [1, 2]
        assert table["material_name"].combine_chunks().to_pylist() == [
            "material_1_updated",
            "material_2",
        ]

    def test_initialize_empty_table(self):
        assert self.db.is_empty()

        files = os.listdir(self.db.db_path)
        assert len(files) == 1

        table = pq.read_table(os.path.join(self.db.db_path, files[0]))
        table_shape = table.shape
        assert table_shape == (0, 1)

        for file in files:
            os.remove(os.path.join(self.db.db_path, file))

        # Test with initial fields
        db = ParquetDB(
            self.db.db_path,
            initial_fields=[
                pa.field("source_id", pa.int64()),
                pa.field("target_id", pa.int64()),
            ],
        )
        assert db.is_empty()

        files = os.listdir(db.db_path)
        assert len(files) == 1

        table = pq.read_table(os.path.join(self.db.db_path, files[0]))
        table_shape = table.shape
        assert table_shape == (0, 3)

    def test_update_multi_keys(self):
        current_data = [
            {"id_1": 100, "id_2": 10, "field_1": "here"},
            {"id_1": 55, "id_2": 11},
            {"id_1": 33, "id_2": 12},
            {"id_1": 12, "id_2": 13},
        ]

        self.db.create(current_data)

        # Create second table with salary data
        incoming_data = [
            {"id_1": 100, "id_2": 10, "field_2": "there"},
            {"id_1": 5, "id_2": 5},
            {"id_1": 33, "id_2": 13},  # Note: emp_id 4 doesn't exist in employees
            {
                "id_1": 33,
                "id_2": 12,
                "field_2": "field_2",
                "field_3": "field_3",
            },  # Note: emp_id 4 doesn't exist in employees
        ]

        incoming_table = ParquetDB.construct_table(incoming_data)

        self.db.update(incoming_table, update_keys=["id_1", "id_2"])

        table = self.db.read()
        print(table.to_pandas())
        assert table["field_1"].combine_chunks().to_pylist() == [
            "here",
            None,
            None,
            None,
        ]
        assert table["field_2"].combine_chunks().to_pylist() == [
            "there",
            None,
            "field_2",
            None,
        ]
        assert table["field_3"].combine_chunks().to_pylist() == [
            None,
            None,
            "field_3",
            None,
        ]

    def test_table_join(self):
        # Create first table with employee data
        current_data = [
            {"id_1": 100, "id_2": 10, "field_1": "here"},
            {"id_1": 33, "id_2": 12},
            {"id_1": 12, "id_2": 13, "field_2": "field_2"},
        ]

        current_table = ParquetDB.construct_table(current_data)

        # Create second table with salary data
        incoming_data = [
            {"id_1": 100, "id_2": 10, "field_2": "there"},
            {"id_1": 5, "id_2": 5},
            {"id_1": 33, "id_2": 13},  # Note: emp_id 4 doesn't exist in employees
            {
                "id_1": 33,
                "id_2": 12,
                "field_2": "field_2",
                "field_3": "field_3",
            },  # Note: emp_id 4 doesn't exist in employees
        ]

        incoming_table = ParquetDB.construct_table(incoming_data)

        join_type = "left outer"
        left_outer_table_pyarrow = incoming_table.join(
            current_table,
            keys=["id_1", "id_2"],
            right_keys=["id_1", "id_2"],
            left_suffix="_incoming",
            right_suffix="_current",
            join_type=join_type,
        )

        column_names = left_outer_table_pyarrow.column_names
        names_sorted = sorted(column_names)
        left_outer_table_pyarrow_sorted = left_outer_table_pyarrow.select(names_sorted)

        left_outer_table = pyarrow_utils.join_tables(
            incoming_table,
            current_table,
            left_keys=["id_1", "id_2"],
            right_keys=["id_1", "id_2"],
            left_suffix="_incoming",
            right_suffix="_current",
            join_type=join_type,
        )

        column_names = left_outer_table.column_names
        names_sorted = sorted(column_names)
        left_outer_table_sorted = left_outer_table.select(names_sorted)

        for name in left_outer_table_sorted.column_names:
            assert (
                left_outer_table_pyarrow_sorted[name].to_pylist()
                == left_outer_table_sorted[name].to_pylist()
            )

        join_type = "right outer"
        right_outer_table_pyarrow = incoming_table.join(
            current_table,
            keys=["id_1", "id_2"],
            right_keys=["id_1", "id_2"],
            left_suffix="_incoming",
            right_suffix="_current",
            join_type=join_type,
        )

        column_names = right_outer_table_pyarrow.column_names
        names_sorted = sorted(column_names)
        right_outer_table_pyarrow_sorted = right_outer_table_pyarrow.select(
            names_sorted
        )

        right_outer_table = pyarrow_utils.join_tables(
            incoming_table,
            current_table,
            left_keys=["id_1", "id_2"],
            right_keys=["id_1", "id_2"],
            left_suffix="_incoming",
            right_suffix="_current",
            join_type=join_type,
        )

        column_names = right_outer_table.column_names
        names_sorted = sorted(column_names)
        right_outer_table_sorted = right_outer_table.select(names_sorted)

        for name in right_outer_table_sorted.column_names:
            assert (
                right_outer_table_pyarrow_sorted[name].to_pylist()
                == right_outer_table_sorted[name].to_pylist()
            )

        join_type = "inner"
        inner_table_pyarrow = incoming_table.join(
            current_table,
            keys=["id_1", "id_2"],
            right_keys=["id_1", "id_2"],
            left_suffix="_incoming",
            right_suffix="_current",
            join_type=join_type,
        )

        column_names = inner_table_pyarrow.column_names
        names_sorted = sorted(column_names)
        inner_table_pyarrow_sorted = inner_table_pyarrow.select(names_sorted)

        inner_table = pyarrow_utils.join_tables(
            incoming_table,
            current_table,
            left_keys=["id_1", "id_2"],
            right_keys=["id_1", "id_2"],
            left_suffix="_incoming",
            right_suffix="_current",
            join_type=join_type,
        )

        column_names = inner_table.column_names
        names_sorted = sorted(column_names)
        inner_table_sorted = inner_table.select(names_sorted)

        for name in inner_table_sorted.column_names:
            assert (
                inner_table_pyarrow_sorted[name].to_pylist()
                == inner_table_sorted[name].to_pylist()
            )

        join_type = "right anti"
        right_anti_table_pyarrow = incoming_table.join(
            current_table,
            keys=["id_1", "id_2"],
            right_keys=["id_1", "id_2"],
            left_suffix="_incoming",
            right_suffix="_current",
            join_type=join_type,
        )

        column_names = right_anti_table_pyarrow.column_names
        names_sorted = sorted(column_names)
        right_anti_table_pyarrow_sorted = right_anti_table_pyarrow.select(names_sorted)

        right_anti_table = pyarrow_utils.join_tables(
            incoming_table,
            current_table,
            left_keys=["id_1", "id_2"],
            right_keys=["id_1", "id_2"],
            left_suffix="_incoming",
            right_suffix="_current",
            join_type=join_type,
        )

        column_names = right_anti_table.column_names
        names_sorted = sorted(column_names)
        right_anti_table_sorted = right_anti_table.select(names_sorted)

        for name in right_anti_table_sorted.column_names:
            assert (
                right_anti_table_pyarrow_sorted[name].to_pylist()
                == right_anti_table_sorted[name].to_pylist()
            )

        join_type = "left anti"
        left_anti_table_pyarrow = incoming_table.join(
            current_table,
            keys=["id_1", "id_2"],
            right_keys=["id_1", "id_2"],
            left_suffix="_incoming",
            right_suffix="_current",
            join_type=join_type,
        )

        column_names = left_anti_table_pyarrow.column_names
        names_sorted = sorted(column_names)
        left_anti_table_pyarrow_sorted = left_anti_table_pyarrow.select(names_sorted)

        left_anti_table = pyarrow_utils.join_tables(
            incoming_table,
            current_table,
            left_keys=["id_1", "id_2"],
            right_keys=["id_1", "id_2"],
            left_suffix="_incoming",
            right_suffix="_current",
            join_type=join_type,
        )

        column_names = left_anti_table.column_names
        names_sorted = sorted(column_names)
        left_anti_table_sorted = left_anti_table.select(names_sorted)

        for name in left_anti_table_sorted.column_names:
            assert (
                left_anti_table_pyarrow_sorted[name].to_pylist()
                == left_anti_table_sorted[name].to_pylist()
            )

        join_type = "left semi"
        left_semi_table_pyarrow = incoming_table.join(
            current_table,
            keys=["id_1", "id_2"],
            right_keys=["id_1", "id_2"],
            left_suffix="_incoming",
            right_suffix="_current",
            join_type=join_type,
        )

        column_names = left_semi_table_pyarrow.column_names
        names_sorted = sorted(column_names)
        left_semi_table_pyarrow_sorted = left_semi_table_pyarrow.select(names_sorted)

        left_semi_table = pyarrow_utils.join_tables(
            incoming_table,
            current_table,
            left_keys=["id_1", "id_2"],
            right_keys=["id_1", "id_2"],
            left_suffix="_incoming",
            right_suffix="_current",
            join_type=join_type,
        )

        column_names = left_semi_table.column_names
        names_sorted = sorted(column_names)
        left_semi_table_sorted = left_semi_table.select(names_sorted)

        for name in left_semi_table_sorted.column_names:
            assert (
                left_semi_table_pyarrow_sorted[name].to_pylist()
                == left_semi_table_sorted[name].to_pylist()
            )

        join_type = "right semi"
        right_semi_table_pyarrow = incoming_table.join(
            current_table,
            keys=["id_1", "id_2"],
            right_keys=["id_1", "id_2"],
            left_suffix="_incoming",
            right_suffix="_current",
            join_type=join_type,
        )

        column_names = right_semi_table_pyarrow.column_names
        names_sorted = sorted(column_names)
        right_semi_table_pyarrow_sorted = right_semi_table_pyarrow.select(names_sorted)

        right_semi_table = pyarrow_utils.join_tables(
            incoming_table,
            current_table,
            left_keys=["id_1", "id_2"],
            right_keys=["id_1", "id_2"],
            left_suffix="_incoming",
            right_suffix="_current",
            join_type=join_type,
        )

        column_names = right_semi_table.column_names
        names_sorted = sorted(column_names)
        right_semi_table_sorted = right_semi_table.select(names_sorted)

        for name in right_semi_table_sorted.column_names:
            assert (
                right_semi_table_pyarrow_sorted[name].to_pylist()
                == right_semi_table_sorted[name].to_pylist()
            )

        join_type = "full outer"
        full_outer_table_pyarrow = incoming_table.join(
            current_table,
            keys=["id_1", "id_2"],
            right_keys=["id_1", "id_2"],
            left_suffix="_incoming",
            right_suffix="_current",
            join_type=join_type,
        )

        column_names = full_outer_table_pyarrow.column_names
        names_sorted = sorted(column_names)
        full_outer_table_pyarrow_sorted = full_outer_table_pyarrow.select(names_sorted)

        full_outer_table = pyarrow_utils.join_tables(
            incoming_table,
            current_table,
            left_keys=["id_1", "id_2"],
            right_keys=["id_1", "id_2"],
            left_suffix="_incoming",
            right_suffix="_current",
            join_type=join_type,
        )

        column_names = full_outer_table.column_names
        names_sorted = sorted(column_names)
        full_outer_table_sorted = full_outer_table.select(names_sorted)

        for name in full_outer_table_sorted.column_names:
            assert (
                full_outer_table_pyarrow_sorted[name].to_pylist()
                == full_outer_table_sorted[name].to_pylist()
            )

    def test_python_objects(self):
        self.db._serialize_python_objects = True
        from pymatgen.core import Structure

        structure = Structure(
            lattice=[[0, 2.13, 2.13], [2.13, 0, 2.13], [2.13, 2.13, 0]],
            species=["Mg", "O"],
            coords=[[0, 0, 0], [0.5, 0.5, 0.5]],
        )
        data = [
            {
                "name": "Alice",
                "age": 30,
                "time": pd.Timestamp("20180310"),
                "structure": None,
            },
            {
                "name": "Alice",
                "age": 30,
                "time": pd.Timestamp("20180310"),
                "structure": structure,
            },
        ]
        self.db.create(data)

        table = self.db.read()
        df = table.to_pandas()

        assert df["structure"][0] is None

        self.db.create(data)

        table = self.db.read()
        df = table.to_pandas()
        assert isinstance(df["structure"][1], Structure)

        data = [
            {"id": 1, "new_field": 1, "structure": None},
            {"id": 0, "new_field": 2, "structure": structure},
        ]
        self.db.update(data)

        table = self.db.read()
        df = table.to_pandas()
        # print(df)
        # print(df["structure"][1])
        assert df["new_field"][1] == 1

        assert table["structure"].type == types.PythonObjectArrowType()


if __name__ == "__main__":
    # unittest.TextTestRunner().run(TestParquetDB('test_nested_data_handling'))
    # unittest.TextTestRunner().run(TestParquetDB('test_update_maintains_existing_extension_arrays'))
    # unittest.TextTestRunner().run(TestParquetDB('test_update_maintains_existing_extension_arrays'))
    # unittest.TextTestRunner().run(TestParquetDB('test_update_maintains_existing_extension_arrays_batches'))
    # unittest.TextTestRunner().run(TestParquetDB('test_metadata'))
    # unittest.TextTestRunner().run(TestParquetDB('test_fixed_shape_tensor'))
    # unittest.TextTestRunner().run(TestParquetDB('test_metadata'))
    # unittest.TextTestRunner().run(TestParquetDB('test_initialize_empty_table'))
    # unittest.TextTestRunner().run(TestParquetDB('test_batch_reading'))
    #  unittest.TextTestRunner().run(TestParquetDB('test_create_and_read'))
    # unittest.TextTestRunner().run(TestParquetDB('test_update_multi_keys'))
    # unittest.TextTestRunner().run(TestParquetDB('test_fixed_shape_tensor'))
    # unittest.TextTestRunner().run(TestParquetDB('test_rename_dataset'))
    # unittest.TextTestRunner().run(TestParquetDB('test_update_with_new_field_included'))
    # unittest.TextTestRunner().run(TestParquetDB("test_python_objects"))
    # unittest.TextTestRunner().run(TestParquetDB('test_update_multi_keys'))
    unittest.main()


# if __name__ == '__main__':
#     for x in range(100):
#         print(f"Iteration {x+1}")

# Create a test suite and add your test case
# suite = unittest.TestLoader().run(TestParquetDB('test_update_multi_keys'))
# Create a test suite and add your test case
# suite = unittest.TestLoader().loadTestsFromTestCase(TestParquetDB)

# Run the tests
# unittest.TextTestRunner().run(suite)
# try:
#     unittest.TextTestRunner().run(TestParquetDB('test_update_multi_keys'))
# except Exception as e:
#     break



import os
import sys
import requests

from dotenv import load_dotenv

load_dotenv()

# Get environment variables
GITHUB_TOKEN = os.getenv('GITHUB_TOKEN')
REPO_NAME = os.getenv('GITHUB_REPOSITORY')
RELEASE_ID = os.getenv('RELEASE_ID')  # Pass the release ID from the GitHub Action context
CHANGELOG = os.getenv('CHANGELOG')

headers = {
    "Authorization": f"token {GITHUB_TOKEN}",
    "Accept": "application/vnd.github.v3+json",
}


def get_release_details(release_id):
    api_url = f"https://api.github.com/repos/{REPO_NAME}/releases/{release_id}"

    response = requests.get(api_url, headers=headers)

    # Check if the request was successful
    if response.status_code == 200:
        release_data = response.json()
        print("Release Data:")
        print(f"Release ID: {release_data['id']}, Tag Name: {release_data['tag_name']}")
        print(f"Release Body: {release_data['body']}")
        print(f"Release URL: {release_data['html_url']}")
    else:
        print(f"Failed to fetch release data: {response.status_code}")
        print(response.json())
        release_data={}
    return release_data



def update_release_body(input_string):
    
    try:
        release_data=get_release_details(RELEASE_ID)
        new_release_body=release_data['body'] + '\n\n' + input_string
    except Exception as e:
        print(f"Error: {e}")
        return None
    
    print(new_release_body)

    api_url = f"https://api.github.com/repos/{REPO_NAME}/releases/{RELEASE_ID}"
    headers = {
        "Authorization": f"token {GITHUB_TOKEN}",
        "Accept": "application/vnd.github.v3+json",
        "Content-Type": "application/json",
    }

    json_data = {
        "body": new_release_body,
    }

    response = requests.patch(api_url, headers=headers, json=json_data)

    # Check if the request was successful

    if response.status_code == 200:
        print("Release updated successfully!")
        print(response.json())  # Optional: Print the updated release details
    else:
        print(f"Failed to update release: {response.status_code}")
        print(response.json())




if __name__ == "__main__":
    # if len(sys.argv) > 1:
    #     input_string = sys.argv[1]
    #     print(f"Received string: {input_string}")
    # else:
    #     input_string=''
    #     print("No string provided.")
    # print(input_string)
    update_release_body(CHANGELOG)



import os
import shutil
import uuid
from filelock import FileLock  # pip install filelock

class TransactionManager:
    def __init__(self, db_path):
        self.db_path = db_path
        self._lock_file = os.path.join(db_path, ".db_lock")
        self._temp_dir = None
        self._active_tx = False
        self._lock = FileLock(self._lock_file)

    def begin_transaction(self):
        # Acquire exclusive lock so no other writes happen
        self._lock.acquire()
        self._active_tx = True
        
        # Create a temp directory for staging any new or updated Parquet files
        self._temp_dir = os.path.join(self.db_path, f"temp_tx_{uuid.uuid4()}")
        os.makedirs(self._temp_dir, exist_ok=True)

    def commit(self):
        if not self._active_tx:
            raise RuntimeError("No active transaction to commit.")

        # Atomically move from temp directory to the main DB folder
        self._apply_changes()

        # Clean up
        self._cleanup()

    def rollback(self):
        if not self._active_tx:
            raise RuntimeError("No active transaction to rollback.")

        # Just discard the temp changes
        self._cleanup()

    def _apply_changes(self):
        """
        Move or rename new Parquet files from the temp directory into the main DB folder 
        in an atomic way. For instance, you might rename or replace the original dataset
        with the newly generated dataset.
        """
        for fname in os.listdir(self._temp_dir):
            src = os.path.join(self._temp_dir, fname)
            dst = os.path.join(self.db_path, fname)
            
            # If you want versioning, you could move the old files to a “backup” or 
            # “versions” folder. For simplicity, we replace them:
            if os.path.exists(dst):
                os.remove(dst)
            
            # Atomic operation: rename in the same filesystem
            os.rename(src, dst)

    def _cleanup(self):
        # Remove the temp directory and release the lock
        if os.path.exists(self._temp_dir):
            shutil.rmtree(self._temp_dir)
        self._active_tx = False
        self._lock.release()



import os
import shutil
import sys
import time
import random
import string
import itertools

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt


import pyarrow as pa
import pyarrow.compute as pc
import pyarrow.dataset as ds
import pyarrow.parquet as pq
import pyarrow.fs as fs

from parquetdb import ParquetDB, config


config.logging_config.loggers.timing.level='ERROR'
config.apply()



def generate_needle_in_haystack_data(n_rows=100, n_columns=100, value=-1):
    random_index = random.randint(0, n_rows-1)
    data=[]

    for j in range(n_rows):
        if j==random_index:
            data.append({f'col_{i}':value for i in range(n_columns)})
        else:
            data.append({f'col_{i}':random.randint(0, 100000) for i in range(n_columns)})
    return data



def benchmark_needle_in_haystack(num_rows):
    db = ParquetDB(dataset_name='parquetdb', dir=benchmark_dir)
    if db.dataset_exists():
        db.drop_dataset()  # Assuming there's a method to clear the database
    data=generate_needle_in_haystack_data(n_rows=num_rows)
    
    start_time=time.time()
    db.create(data)
    insert_time = time.time() - start_time
    data=None
    start_time=time.time()
    data = db.read(columns=['col_0'],filters=[(pc.field('col_0') == -1)]).combine_chunks()
    assert data['col_0'].to_pylist()[0]==-1
    read_time = time.time() - start_time
    return insert_time, read_time




if __name__ == '__main__':
    save_dir=os.path.join(config.data_dir, 'benchmarks', 'parquetdb')
    os.makedirs(save_dir, exist_ok=True)

    benchmark_dir=os.path.join(save_dir, 'BenchmarkDB')
    dataset_dir=os.path.join(benchmark_dir, 'parquetdb')

    if os.path.exists(dataset_dir):
        shutil.rmtree(dataset_dir)


    db = ParquetDB(dataset_name='parquetdb', dir=benchmark_dir)


    row_counts = [1, 10, 100, 1000, 10000, 100000, 1000000]
    insertion_times = []
    reading_times = []


    results={
        'create_times':[],
        'read_times':[],
        'n_rows':[]
    }
    for num_rows in row_counts:
        print(f'Benchmarking {num_rows} rows...')
        
        insert_time,read_time = benchmark_needle_in_haystack(num_rows)
        insertion_times.append(insert_time)
        reading_times.append(read_time)
        
        print(f'Insertion Time for {num_rows} rows: {insert_time:.4f} seconds')
        print(f'Reading Time for {num_rows} rows: {read_time:.4f} seconds')
        print('---')
        
    results['create_times']=insertion_times
    results['read_times']=reading_times
    results['n_rows']=row_counts

    df=pd.DataFrame(results)
    df.to_csv(os.path.join(save_dir, 'parquetdb_needle_in_haystack_benchmark.csv'), index=False)


import logging
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor
from functools import partial

from dask.distributed import Client
from pathos.pools import ProcessPool
from pathos.threading import ThreadPool

from parquetdb.utils.config import config

logger = logging.getLogger(__name__)

# def parallel_apply(func, data, executor=None):

#     if executor is None:
#         executor=ThreadPoolExecutor()
#     with executor:
#         future = executor.submit(func, data)
#         return future.result()


# def parallel_apply(func, data, executor=None):

#     if executor is None:
#         executor = ProcessPool()

#     if len(data) > 3500:
#         with executor:
#             future = executor.map(func, data)
#             return future
#     else:
#         results = [func(item) for item in data]
#         return results


def parallel_apply(func, data, processes=False):
    if len(data) > 2000 and config.use_multiprocessing:
        with Client(
            silence_logs=logging.ERROR,
            processes=processes,
        ) as client:
            serialized_futures = client.map(func, data)
            results = client.gather(serialized_futures)
    else:
        results = [func(item) for item in data]
    return results


def mp_task(func, list, n_cores=None, **kwargs):
    """
    Processes tasks in parallel using a pool of worker processes.

    This function applies a given function to a list of items in parallel, using
    multiprocessing with a specified number of cores. Each item in the list is processed
    by the function, and additional arguments can be passed through `kwargs`.
    By defualt, it will detect the number of cores available unless specified otherwise.

    Parameters
    ----------
    func : Callable
        The function to be applied to each item in the list.
    list : list
        A list of items to be processed by the function.
    n_cores : int, optional
        The number of cores to use for multiprocessing (default is None).
    **kwargs
        Additional keyword arguments to be passed to `func`.

    Returns
    -------
    list
        A list of results obtained by applying `func` to each item in the input list.
    """
    if n_cores is None:
        executor = ProcessPool()
    else:
        executor = ProcessPool(n_cores)
    try:
        if len(list) > 3500:
            with executor:
                results = executor.map(partial(func, **kwargs), list)
        else:
            results = [func(item, **kwargs) for item in list]
        logger.info("Tasks processed successfully.")

        return results
    except:
        logging.exception("Error processing tasks in parallel.")



import os
import sys
import subprocess


from dotenv import load_dotenv
from datetime import datetime

import requests

load_dotenv()

api_key = os.getenv("OPENAI_API_KEY")

GITHUB_TOKEN = os.getenv("GITHUB_TOKEN")
REPO_NAME = os.getenv("REPO_NAME")
TAG = os.getenv('TAG')

def run_git_command(command):
    try:
        result = subprocess.run(['git'] + command.split(), 
                                check=True, 
                                capture_output=True, 
                                text=True)
        return result.stdout
    except subprocess.CalledProcessError as e:
        print(f"An error occurred: {e}")
        return e.stderr
    
def bash_command(command):
    try:
        result = subprocess.run(command.split(), 
                                check=True, 
                                capture_output=True, 
                                text=True)
        return result.stdout
    except subprocess.CalledProcessError as e:
        print(f"An error occurred: {e}")
        return e.stderr

def generate_changelog_message():
    # Example usage:
    bash_command('git fetch --all --tags')
    current_version = bash_command('git tag -l --sort=v:refname').strip()
    print(f"Current Version: {current_version}")
    print(f'git log --pretty=format:"%h-%s" {current_version}..')


headers = {
    "Authorization": f"token {GITHUB_TOKEN}",
    "Accept": "application/vnd.github.v3+json",
}


def get_release_ids():
    

    api_url = f"https://api.github.com/repos/{REPO_NAME}/releases"
    print(api_url)
    response = requests.get(api_url, headers=headers)

    # Check if the request was successful
    if response.status_code == 200:
        release_data = response.json()
        print("Release Data:")
        for release in release_data:
            print(f"Release ID: {release['id']}, Tag Name: {release['tag_name']}")
    else:
        print(f"Failed to fetch release data: {response.status_code}")
        print(response.json())

        
if __name__ == "__main__":

    get_release_ids()
    # generate_changelog_message()


import json
import os
import shutil
import sys
import time
import random
import string
import itertools

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt


import pyarrow as pa
import pyarrow.compute as pc
import pyarrow.dataset as ds
import pyarrow.parquet as pq
import pyarrow.fs as fs

from parquetdb import ParquetDB, config
from parquetdb.utils import general_utils

config.logging_config.loggers.timing.level='ERROR'
config.apply()


if __name__ == '__main__':
    save_dir=os.path.join(config.data_dir, 'benchmarks', 'parquetdb')
    os.makedirs(save_dir, exist_ok=True)

    benchmark_dir=os.path.join(save_dir, 'BenchmarkDB')
    dataset_dir=os.path.join(save_dir, 'parquetdb')

    if os.path.exists(dataset_dir):
        shutil.rmtree(dataset_dir)


    db = ParquetDB(dataset_name='parquetdb', dir=benchmark_dir)
    
    if db.dataset_exists():
        db.drop_dataset()  # Assuming there's a method to clear the database
    data=general_utils.generate_pandas_data(n_rows=1000000)
    
    start_time=time.time()
    db.create(data)
    insert_time = time.time() - start_time
    del data
    
    start_time=time.time()
    data = db.read()
    read_time = time.time() - start_time
    del data


    benchmark_dict={}
    new_data=general_utils.generate_pylist_data(n_rows=10, n_columns=110)
    
    start_time=time.time()
    db.create(new_data)
    
    benchmark_dict['create_new_schema_time']=time.time() - start_time
    benchmark_dict['new_fields_added']=10
    benchmark_dict['new_rows_added']=10
    
        
   
    print('---')
    with open(os.path.join(benchmark_dir, 'schema_change_benchmark.json'), 'w') as f:
        json.dump(benchmark_dict, f, indent=4)


import os
import shutil
import sys
import time
import random
import string
import itertools

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt


import pyarrow as pa
import pyarrow.compute as pc
import pyarrow.dataset as ds
import pyarrow.parquet as pq
import pyarrow.fs as fs

from parquetdb import ParquetDB, config


config.logging_config.loggers.timing.level='ERROR'
config.apply()


def get_byte_size_of_type(data_type, n_char=100):
    if data_type == int:
        value = random.randint(0, 100000)
    elif data_type == float:
        value = random.uniform(0, 100000)
    elif data_type == str:
        value = ''.join(random.choices(string.ascii_letters, k=n_char))
    elif data_type == bool:
        value = random.choice([True, False])
    return sys.getsizeof(value)

def chunk_list_iter(lst, chunk_size):
    it = iter(lst)
    return iter(lambda: list(itertools.islice(it, chunk_size)), [])

# Function to generate random data for a given number of rows and columns
def generate_random_data(num_rows, num_cols):
    data_types = [int, float, str, bool]
    data = []
    for _ in range(num_rows):
        row = {}
        for col_index in range(num_cols):
            data_type = random.choice(data_types)
            col_name = f'col_{col_index}_{data_type.__name__}'
            if data_type == int:
                value = random.randint(0, 100000)
            elif data_type == float:
                value = random.uniform(0, 100000)
            elif data_type == str:
                value = ''.join(random.choices(string.ascii_letters, k=100))
            elif data_type == bool:
                value = random.choice([True, False])
            row[col_name] = value
        data.append(row)
    return data


def create_data(n_rows=100, n_columns=1000, data_type=int, n_char=100):
    
    if data_type == int:
        value = random.randint(0, 100000)
    elif data_type == float:
        value = random.uniform(0, 100000)
    elif data_type == str:
        value = ''.join(random.choices(string.ascii_letters, k=n_char))
    elif data_type == bool:
        value = random.choice([True, False])
        
    val=1000000
    data=[]
    for _ in range(n_rows):
        data.append({f'col_{i}':value for i in range(0,n_columns)})
    return data


def measure_memory(n_rows=100, n_columns=10, data_type=int):
    data=create_data(n_rows=n_rows, n_columns=n_columns, data_type=data_type)
    mem_before=pa.total_allocated_bytes()
    table=pa.Table.from_pylist(data)
    mem_after=pa.total_allocated_bytes()-mem_before
    return mem_after


# Function to benchmark the create operation
def benchmark_create(db, data, multiple_creates=False, batch_size=1000):
    start_time = time.time()
    if multiple_creates:
        # Split data into batches and create in multiple operations
        for i in range(0, len(data), batch_size):
            batch = data[i:i+batch_size]
            db.create(batch)
    else:
        # Single create operation
        db.create(data)
    end_time = time.time()
    return end_time - start_time

# Initialize the database


if __name__ == '__main__':
    save_dir=os.path.join(config.data_dir, 'benchmarks', 'parquetdb')
    benchmark_dir=os.path.join(save_dir, 'BenchmarkDB')
    
    dataset_dir=os.path.join(benchmark_dir, 'parquetdb')
    if os.path.exists(dataset_dir):
        shutil.rmtree(dataset_dir)

    db = ParquetDB(dataset_name='parquetdb', dir=benchmark_dir)

    # Parameters for the experiment
    num_columns_list = [1, 10, 100]
    num_rows_list = [1, 10, 100, 1000, 10000, 100000]
    multiple_creates_options = [False, True]
    batch_size_list  = [1000, 10000, 100000]
    data_types=[int,float,str,bool]
    n_chars=np.arange(1,10)
    # n_chars=[1,10,100]

    results = []
    # Run the experiment
    for num_cols in num_columns_list:
        for num_rows in num_rows_list:
            for data_type in data_types:
                if data_type==str:
                    n_chars_init=n_chars
                else:
                    n_chars_init=[1]
                
                for n_char in n_chars_init:
                    
                    byte_size=get_byte_size_of_type(data_type, n_char=n_char)
                    data = create_data(n_rows=num_rows, n_columns=num_cols, data_type=data_type, n_char=n_char)
                    # data = generate_random_data(num_rows, num_cols)
                    for multiple_creates in multiple_creates_options:
                        if multiple_creates:
                            batch_size_list_init=batch_size_list
                        else:
                            batch_size_list_init=[len(data)]
                        for batch_size in batch_size_list_init:
                            params_dict={
                                'num_cols': num_cols,
                                'num_rows': num_rows,
                                'n_chars': n_char,
                                'data_type': data_type,
                                'byte_size': byte_size,
                                'multiple_creates': multiple_creates,
                                'batch_size': batch_size,
                            }
                            if db.dataset_exists():
                                db.drop_dataset()  # Assuming there's a method to clear the database
                            time_taken = benchmark_create(db, data, multiple_creates=multiple_creates, batch_size=batch_size)
                            
                            params_dict.update({'time_taken': time_taken})
                            results.append(params_dict)
                            print(f'Columns: {num_cols}, Rows: {num_rows}, '
                                f'Multiple Creates: {multiple_creates}, Batch size: {batch_size}, '
                                f'Data Type: {data_type}, N Chars: {n_char}, '
                                f'Time Taken: {time_taken:.2f} seconds ')

    # Convert results to DataFrame
    results_df = pd.DataFrame(results)
    results_df.to_csv(os.path.join(save_dir, 'experiment_benchmark.csv'), index=False)



import json
import os
import logging
from glob import glob
import shutil
import time

import numpy as np
from pyarrow import compute as pc

from parquetdb.core.parquetdb import LoadConfig, NormalizeConfig
from parquetdb.utils.general_utils import timeit
from parquetdb import ParquetDB, config
from parquetdb.utils.external_utils import download_alexandria_3d_database

import matplotlib.pyplot as plt

config.logging_config.loggers.parquetdb.level='ERROR'
config.apply()

def main():
    
    base_dir=os.path.join('data','external', 'alexandria', 'AlexandriaDB')
    benchmark_dir=os.path.join(config.data_dir, 'benchmarks', 'alexandria')
    
    from_scratch=False

    os.makedirs(benchmark_dir, exist_ok=True)
    
    ################################################################################################
    
    alexandria_dir=download_alexandria_database(base_dir, from_scratch=from_scratch)

    ################################################################################################
    dataset_name = ['alexandria_3D', 'alexandria_3D_test', 'alexandria_3D_benchmark'][0]
    
    db=ParquetDB(dataset_name=dataset_name,dir=base_dir)
    
    benchmark_dict={
        'create_times':[],
        'json_load_times':[],
        'n_rows_per_file':[],
    }
    
    task_benchmark_dict={
        'task_names':[],
        'task_times':[]
    }
        
    ################################################################################################

    # json_load_times, create_times, n_materials_per_file= create_database_if_empty(db, alexandria_dir)
            
    # benchmark_dict['create_times']=create_times
    # benchmark_dict['json_load_times']=json_load_times
    # benchmark_dict['n_rows_per_file']=n_materials_per_file
    
    # with open(os.path.join(benchmark_dir, f"create_load_benchmark_{dataset_name}.json"), 'w') as f:
    #     json.dump(benchmark_dict, f, indent=4)

    ################################################################################################
    # Now that the data is in the database, we can normalize it. 
    # This means we enfore our parquet files have a certain number of rows. 
    # This improve performance as there are less files to io operations in subsequent steps.
    
    start_time = time.time()
    task_name=normalize_dataset(db)
    task_benchmark_dict['task_times'].append(time.time() - start_time)
    task_benchmark_dict['task_names'].append(task_name)

    ################################################################################################
    # Here we read the single column 'id' and track the time taken
    
    start_time=time.time()
    task_name=read_single_column(db)
    task_benchmark_dict['task_times'].append(time.time() - start_time)
    task_benchmark_dict['task_names'].append(task_name)
    
    ################################################################################################
    # Here, we test reading specific ids and track the time taken
    
    start_time=time.time()
    task_name=read_specific_ids(db)
    task_benchmark_dict['task_times'].append(time.time() - start_time)
    task_benchmark_dict['task_names'].append(task_name)
    
    ################################################################################################
    # Here, we read the energy column and calculate the min and max and track the time taken

    start_time=time.time()
    task_name=read_energy_min_max(db)
    task_benchmark_dict['task_times'].append(time.time() - start_time)
    task_benchmark_dict['task_names'].append(task_name)
    
    ################################################################################################
    # Here, we read the records filtered by energy above -1.0 and track the time taken
    
    start_time=time.time()
    task_name=read_filtered_energy_above_minus_one(db)
    task_benchmark_dict['task_times'].append(time.time() - start_time)
    task_benchmark_dict['task_names'].append(task_name)
    
    ################################################################################################
    # Here, we read the records filtered by space group 204 and track the time taken

    start_time=time.time()
    task_name=read_filtered_spg_204(db)
    task_benchmark_dict['task_times'].append(time.time() - start_time)
    task_benchmark_dict['task_names'].append(task_name)
    
    ################################################################################################
    # Here, we read the records filtered by space group 204 in batches and track the time taken
    
    start_time=time.time()
    task_name=read_filtered_spg_batches(db)
    task_benchmark_dict['task_times'].append(time.time() - start_time)
    task_benchmark_dict['task_names'].append(task_name)

    
    ################################################################################################
    # Here, we read the nested column selection and track the time taken
    
    start_time=time.time()
    task_name=read_nested_column_selection(db)
    task_benchmark_dict['task_times'].append(time.time() - start_time)
    task_benchmark_dict['task_names'].append(task_name)

    
    ################################################################################################
    # Here, we read the nested structure into a class and track the time taken
    
    start_time=time.time()
    task_name=read_nested_structure_into_class_from_scratch(db)
    task_benchmark_dict['task_times'].append(time.time() - start_time)
    task_benchmark_dict['task_names'].append(task_name)
        
    
    ################################################################################################
    # Here, we read the nested structure into a class and track the time taken
    
    start_time=time.time()
    task_name=read_nested_structure_into_class(db)
    task_benchmark_dict['task_times'].append(time.time() - start_time)
    task_benchmark_dict['task_names'].append(task_name)
        
    ################################################################################################
    # Here, we update a single record and track the time taken
    
    start_time = time.time()
    task_name=updating_single_record(db)
    task_benchmark_dict['task_times'].append(time.time() - start_time)
    task_benchmark_dict['task_names'].append(task_name)
    
    ################################################################################################
    # Here, we update a single record and track the time taken
    
    task_name,final_time=updating_100000_records(db)
    task_benchmark_dict['task_times'].append(final_time)
    task_benchmark_dict['task_names'].append(task_name)
    
    ################################################################################################
    # Here, we read all the lattice matrix for space group 204 into a numpy array and track the time taken
    
    start_time=time.time()
    task_name=read_lattice_matrix_spg_204(db)
    task_benchmark_dict['task_times'].append(time.time() - start_time)
    task_benchmark_dict['task_names'].append(task_name)
    
    ################################################################################################
    
    with open(os.path.join(benchmark_dir, f"task_benchmark_{dataset_name}.json"), 'w') as f:
        json.dump(task_benchmark_dict, f, indent=4)
    

def download_alexandria_database(base_dir, from_scratch=False):
    print("Starting task: download_alexandria_database")
    if from_scratch and os.path.exists(base_dir):
        print(f"Removing existing directory: {base_dir}")
        shutil.rmtree(base_dir, ignore_errors=True)
        
    # Here we download the database and save it to the data directory
    output_dir=os.path.join(config.data_dir, 'external', 'alexandria')
    alexandria_dir = download_alexandria_3d_database(output_dir, n_cores=8)
    print("Done with task: download_alexandria_database")
    print('-'*200)
    return alexandria_dir
    
def create_database_if_empty(db, alexandria_dir):
    print("Starting task: create_database_if_empty")
    
    start_time = time.time()
    
    json_load_times=[]
    create_times=[]
    n_materials_per_file=[]
    if len(os.listdir(db.dataset_dir))==0:
        print("The dataset does not exist. Creating it.")
        json_files=glob(os.path.join(alexandria_dir,'*.json'))
        for json_file in json_files[:]:
            
            start_time = time.time()
            with open(json_file, 'r') as f:
                data = json.load(f)
            json_load_time = time.time() - start_time
            
            base_name=os.path.basename(json_file)
            n_materials=len(data['entries'])
            print(f"Processing file: {base_name}")
            print(f"Number of materials: {n_materials}")
            try:
                # Since we are importing alot of data it is best
                # to normalize the database afterwards
                start_time = time.time()
                db.create(data['entries'], normalize_dataset=False)
                create_time = time.time() - start_time
                
                create_times.append(create_time)
                n_materials_per_file.append(n_materials)
                json_load_times.append(json_load_time)
                
            except Exception as e:
                print(e)
            
            
            print(f"Time taken to create dataset: {time.time() - start_time}")
            print('-'*100)

    print("Done with task: create_database_if_empty")
    print('-'*200)
    return json_load_times, create_times, n_materials_per_file

def normalize_dataset(db):
    task_name='normalize_dataset'
    print("Starting task: normalize_dataset")
    normalize_config= NormalizeConfig(
        load_format='batches',      # Uses the batch generator to normalize
        batch_readahead=10,         # Controls the number of batches to load in memory a head of time. This will have impacts on amount of RAM consumed
        fragment_readahead=2,       # Controls the number of files to load in memory ahead of time. This will have impacts on amount of RAM consumed
        batch_size = 100000,        # Controls the batchsize when to use when normalizing. This will have impacts on amount of RAM consumed
        max_rows_per_file=500000,   # Controls the max number of rows per parquet file
        max_rows_per_group=500000)  # Controls the max number of rows per group parquet file
    db.normalize(normalize_config=normalize_config)
    print("Done with task: normalize_dataset")
    print('-'*200)
    return task_name

def read_single_column(db):
    task_name='read_single_column'
    print("Starting task: read_single_column")
    table = db.read(
                        columns=['id'],
                        load_format='table')

    print(table.shape)
    print('-'*200)
    return task_name

def read_specific_ids(db):
    task_name='read_specific_ids'
    print("Starting task: read_specific_ids")
    table = db.read(
                    ids=[0,10,100,1000,10000,100000,1000000], # Controls which rows we want to read 
                    load_format='table' # Controls the output format. The options are 'table', 'batches', `dataset`.
                    )

    df = table.to_pandas() # Converts the table to a pandas dataframe
    print(df['id'])
    print(df.head())
    print(df.shape)

    print(f"Data : {df.iloc[0]['data.spg']}")
    print(list(df.columns))
    print("Done with task: read_specific_ids")
    print('-'*200)
    return task_name

def read_energy_min_max(db):
    task_name='read_energy_min_max'
    print("Starting task: read_energy_min_max")
    table = db.read(
                        columns=['energy'],
                        load_format='table')
    print(table.shape)
    
    result = pc.min_max(table['energy'])
    # The result will be a struct with 'min' and 'max' fields
    min_value = result['min'].as_py()
    max_value = result['max'].as_py()

    print(f"Min: {min_value}, Max: {max_value}")
    print("Done with task: read_energy_min_max") 
    print('-'*200)
    return task_name
        
def read_filtered_energy_above_minus_one(db):
    task_name='read_filtered_energy_above_-1'
    
    """Read records filtered by energy above -1.0 and track timing."""
    print("Starting task: read_filtered_energy_above_minus_one")

    table = db.read(
                    columns=['id','energy'],
                    filters=[pc.field('energy') > -1.0],
                    load_format='table')
    
    df = table.to_pandas()  # Converts the table to a pandas dataframe
    print(df.head())
    print(df.shape)
    print("Done with task: read_filtered_energy_above_minus_one")
    print('-'*200)
    return task_name
        
def read_filtered_spg_204(db):
    task_name='read_filtered_spg_204_table'

    print("Starting task: read_filtered_spg_204")
    table = db.read(
                    columns=['id', 'data.spg'],
                    filters=[pc.field('data.spg') == 204],
                    load_format='table')
    
    df = table.to_pandas()  # Converts the table to a pandas dataframe
    print(df.head())
    print(df.shape)
    
    print("Done with task: read_filtered_spg_204")
    print('-'*200)
    return task_name

def read_filtered_spg_batches(db):
    task_name='read_filtered_spg_batches'
    print("Starting task: read_filtered_spg_batches")
    generator = db.read(
                        load_format='batches', 
                        batch_size=1000,
                        load_config=LoadConfig(batch_readahead=10,
                                                fragment_readahead=2,
                                                fragment_scan_options=None,
                                                use_threads=True,
                                                memory_pool=None),
                        columns=['id', 'data.spg'],
                        filters=[pc.field('data.spg') == 204])

    batch_count = 0
    num_rows = 0
    for table in generator:
        df = table.to_pandas()
        num_rows += table.num_rows
        batch_count += 1
    print(f"Total number of rows: {num_rows}, Batches: {batch_count}")
    print("Done with task: read_filtered_spg_batches")
    print('-'*200)
    return task_name

def read_nested_column_selection(db):
    task_name='read_nested_column_selection'
    
    print("Starting task: read_nested_column_selection")
    table = db.read(
                    columns=['id', 'structure.sites'],
                    load_format='table')

    print(table.shape)
    print(table['structure.sites'].type)
    print(table['structure.sites'].combine_chunks().type)
    print("Done with task: read_nested_column_selection")
    print('-'*200)
    return task_name

def read_nested_structure_into_class_from_scratch(db):
    # By default the database flattens nested structure for storage. 
    # However, we provide an option to rebuild the nested structure. This will create a new dataset in {dataset_name}_nested.
    # After the creation of the new dataset, the query parameters are applied to the new dataset.
    task_name='read_nested_structure_into_class_from_scratch'
    
    print("Starting task: read_nested_structure_into_class_from_scratch")
    table = db.read(columns=['id', 'structure','data'], # Instead of using the flatten syntax, we can use the nested syntax
                    ids=[0,1000000],
                    load_format='table',
                    rebuild_nested_struct=True,         # When set to True to rebuild the nested structure
                    rebuild_nested_from_scratch=True,  # When set to True, the nested structure will be rebuilt from scratch
                    normalize_config=NormalizeConfig(
                        load_format='batches',
                        batch_readahead=2,
                        fragment_readahead=2,
                        batch_size=50000,
                        max_rows_per_file=500000,
                        min_rows_per_group=0,
                        max_rows_per_group=500000
                    )
                    )

    print(table.shape)
    print(table['data'].type)
    
    print("structure type")
    print(table['structure'].type)
    try:
        from pymatgen.core.structure import Structure
        
        structure=Structure.from_dict(table['structure'].combine_chunks().to_pylist()[0])
        
        print(structure)
    except Exception as e:
        print(e)
    print("Done with task: read_nested_structure_into_class_from_scratch")
    print('-'*200)
    return task_name


def read_nested_structure_into_class(db):
    # By default the database flattens nested structure for storage. 
    # However, we provide an option to rebuild the nested structure. This will create a new dataset in {dataset_name}_nested.
    # After the creation of the new dataset, the query parameters are applied to the new dataset.
    task_name='read_nested_structure_into_class'
    
    print("Starting task: read_nested_structure_into_class")
    table = db.read(columns=['id', 'structure','data'], # Instead of using the flatten syntax, we can use the nested syntax
                    ids=[0,1000000],
                    load_format='table',
                    rebuild_nested_struct=True,         # When set to True to rebuild the nested structure
                    rebuild_nested_from_scratch=False,  # When set to True, the nested structure will be rebuilt from scratch
                    normalize_config=NormalizeConfig(
                        load_format='batches',
                        batch_readahead=2,
                        fragment_readahead=2,
                        batch_size=50000,
                        max_rows_per_file=500000,
                        min_rows_per_group=0,
                        max_rows_per_group=500000
                    )
                    )

    print(table.shape)
    print(table['data'].type)
    
    print("structure type")
    print(table['structure'].type)
    try:
        from pymatgen.core.structure import Structure
        
        structure=Structure.from_dict(table['structure'].combine_chunks().to_pylist()[0])
        
        print(structure)
    except Exception as e:
        print(e)
    print("Done with task: read_nested_structure_into_class")
    print('-'*200)
    return task_name
    
def updating_single_record(db):
    print("Starting task: updating_single_record")
    task_name='updating_single_record'
    
    # table=db.read(ids=[0])
    # df=table.to_pandas()
    # print(df['data.spg'])
    start_time = time.time()
    db.update([{'id':0, 'data.spg':210}], 
            normalize_config=NormalizeConfig(
                load_format='batches',      # Uses the batch generator to normalize
                batch_readahead=10,   # Controls the number of batches to load in memory a head of time. This will have impacts on amount of RAM consumed
                fragment_readahead=2,  # Controls the number of files to load in memory ahead of time. This will have impacts on amount of RAM consumed
                batch_size = 100000,       # Controls the batchsize when to use when normalizing. This will have impacts on amount of RAM consumed
                max_rows_per_file=500000,  # Controls the max number of rows per parquet file                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          
                max_rows_per_group=500000) # Controls the max number of rows per group parquet file
          )
    print(f"Time taken to update: {time.time() - start_time}")
    table=db.read(ids=[0])
    print(table.shape)
    print("Done with task: updating_dataset")
    print('-'*200)
    return task_name
    
def updating_100000_records(db):
    print("Starting task: updating_100000_records")
    task_name='updating_100000_records'
    
    # Read first 100000 records
    table = db.read(columns=['id'])
    # df = table.to_pandas()
    # print("Original spg values for first few records:")
    # print(df['data.spg'].head())

    # Create list of 100000 updates
    updates = [{'id': np.random.randint(0, table.num_rows), 'data.spg': 210} for _ in range(100000)]
    start_time = time.time()
    db.update(updates,
            normalize_config=NormalizeConfig(
                load_format='batches',      # Uses the batch generator to normalize
                batch_readahead=10,   # Controls the number of batches to load in memory a head of time. This will have impacts on amount of RAM consumed
                fragment_readahead=2,  # Controls the number of files to load in memory ahead of time. This will have impacts on amount of RAM consumed
                batch_size = 100000,       # Controls the batchsize when to use when normalizing. This will have impacts on amount of RAM consumed
                max_rows_per_file=500000,  # Controls the max number of rows per parquet file
                max_rows_per_group=500000) # Controls the max number of rows per group parquet file
          )
    final_time = time.time() - start_time
    # Verify updates
    print("Updated spg values for first few records:")
    print(table.shape)
    # print(df['data.spg'].head())
    print("Done with task: updating_100000_records")
    print('-'*200)
    return task_name,final_time
    
def read_lattice_matrix_spg_204(db):
    task_name='read_lattice_matrix_spg_204'
    
    print("Starting task: read_lattice_matrix_spg_204")
    
    table = db.read(columns=['structure.lattice.matrix'], 
                    filters=[pc.field('data.spg') == 204])
    lattice = table['structure.lattice.matrix'].combine_chunks().to_numpy_ndarray()
    print(lattice.shape)
    print("Done with task: read_lattice_matrix_spg_204")
    print('-'*200)
    return task_name


    
if __name__ == "__main__":
    main()



import os
import random
import sqlite3
import time

import pandas as pd
from parquetdb import config

def generate_data(num_rows, n_columns=100):
    data = []
    for i in range(num_rows):
        # Include a unique identifier 'id' as the first column
        row = (i,) + tuple(random.randint(0, 1000000) for _ in range(n_columns - 1))
        data.append(row)
    return data

def remove_db_file(db_name):
    if os.path.exists(db_name):
        os.remove(db_name)





def benchmark_read_write_update(num_rows, db_name='benchmark.db'):
    
    #####################################################################
    # Reconnect to simulate fresh read from disk
    

    # Prepare update data (new values for existing IDs)
    update_data = generate_data(num_rows)
    
    start_time = time.time()
    update_values = []
    for row in update_data:
        # All columns except 'id', then 'id' for WHERE clause
        update_values.append(row[1:] + (row[0],))

    # Measure update time
    
    
    conn = sqlite3.connect(db_name)
    
    # Construct update query to update columns col1 to col99
    update_query = f"UPDATE test_table SET " + ', '.join([f'col{i} = ?' for i in range(1, 100)]) + " WHERE col0 = ?"
    print(f"Data Query Preperation Time: {time.time() - start_time}")
    
    start_time = time.time()
    cursor = conn.cursor()
    cursor.executemany(update_query, update_values)
    conn.commit()
    conn.close()
    update_time = time.time() - start_time
    
    update_data=None

    return insert_time, update_time

if __name__ == '__main__':
    save_dir = os.path.join(config.data_dir, 'benchmarks', 'sqlite')
    os.makedirs(save_dir, exist_ok=True)
    db_name = os.path.join(save_dir, 'benchmark.db')
    
    # Remove existing database file to start fresh
    remove_db_file(db_name)

    ############################################################################################################################
    # Generate data
    data = generate_data(num_rows=1000000)
    # Connect to SQLite database on disk
    conn = sqlite3.connect(db_name)
    cursor = conn.cursor()

    # Create table with 'id' as the primary key
    columns = ', '.join([f'col{i} INTEGER' for i in range(0, 100)])
    cursor.execute(f'CREATE TABLE IF NOT EXISTS test_table ({columns})')

    # Insert data
    placeholders = ', '.join('?' for _ in range(100))
    insert_query = f'INSERT INTO test_table VALUES ({placeholders})'
    conn.execute('PRAGMA synchronous = OFF')
    conn.execute('PRAGMA journal_mode = MEMORY')

    start_time = time.time()
    cursor.executemany(insert_query, data)
    conn.commit()
    insert_time = time.time() - start_time

    # Ensure data is written to disk
    conn.close()
    data=None

    ########################################################################

    conn = sqlite3.connect(db_name)
    cursor = conn.cursor()

    cursor.execute('CREATE INDEX idx_col0 ON test_table (col0)')
    conn.commit()
    # Ensure data is written to disk
    conn.close()
    ############################################################################################################################
    
    
    
    row_counts = [1, 10, 100, 1000, 10000, 100000, 1000000]
    insertion_times = []
    reading_times = []
    update_times = []

    results = {
        'create_times': [],
        'update_times': [],
        'n_rows': []
    }
    

    for num_rows in row_counts:
        print(f'Benchmarking {num_rows} rows...')
        
        insert_time, update_time = benchmark_read_write_update(num_rows, db_name=db_name)
        insertion_times.append(insert_time)

        update_times.append(update_time)
        
        print(f'Insertion Time for {num_rows} rows: {insert_time:.4f} seconds')
        print(f'Update Time for {num_rows} rows: {update_time:.4f} seconds')
        print('---')
        
    results['create_times'] = insertion_times
    # results['read_times'] = reading_times
    results['update_times'] = update_times
    results['n_rows'] = row_counts

    # df = pd.DataFrame(results)
    # df.to_csv(os.path.join(save_dir, 'sqlite_update_into_constant_rows_benchmark.csv'), index=False)
    
    
    # Remove existing database file to start fresh
    remove_db_file(db_name)
    


import os
import requests

# Your GitHub token
GITHUB_TOKEN = os.getenv('GITHUB_TOKEN')
REPO_NAME = os.getenv('GITHUB_REPOSITORY')
RELEASE_ID = os.getenv('RELEASE_ID')  # Pass the release ID from the GitHub Action context
TAG = os.getenv('TAG')

# GitHub API base URL for releases
api_base_url = f"https://api.github.com/repos/{REPO_NAME}"

# Headers for authentication
headers = {
    "Authorization": f"token {GITHUB_TOKEN}",
    "Accept": "application/vnd.github.v3+json",
}

# Step 1: Delete the release using the release ID
def delete_release(release_id):
    if release_id:
        api_url = f"{api_base_url}/releases/{release_id}"
        response = requests.delete(api_url, headers=headers)
        if response.status_code == 204:
            print(f"Release with ID {release_id} deleted successfully.")
        else:
            print(f"Failed to delete release: {response.status_code}")
            print(response.json())

# Step 2: Optionally delete the tag associated with the release
def delete_tag(tag_name):
    api_url = f"{api_base_url}/git/refs/tags/{tag_name}"
    response = requests.delete(api_url, headers=headers)
    if response.status_code == 204:
        print(f"Tag {tag_name} deleted successfully.")
    else:
        print(f"Failed to delete tag {tag_name}: {response.status_code}")
        print(response.json())

# Main logic
if __name__ == "__main__":

    # Delete the release
    delete_release(RELEASE_ID)
    
    # Optionally delete the tag
    delete_tag(TAG)



import os
import shutil
import sys
import time
import random
import string
import itertools

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt


import pyarrow as pa
import pyarrow.compute as pc
import pyarrow.dataset as ds
import pyarrow.parquet as pq
import pyarrow.fs as fs

from parquetdb import ParquetDB, config


config.logging_config.loggers.timing.level='ERROR'
config.apply()

def generate_data(n_rows=100, n_columns=100):
    data=[]
    for _ in range(n_rows):
        data.append({f'col_{i}':random.randint(0, 100000) for i in range(0,n_columns)})
    
    return data


def generate_update_data(n_rows=100, n_columns=100):
    data = []
    for i in range(n_rows):
        row = {'id': i}  # Unique identifier for each row
        row.update({f'col_{j}': random.randint(0, 100000) for j in range(n_columns)})
        data.append(row)
    return data

def generate_pyarrow_data(num_columns=100, num_rows=1000, min_value=0, max_value=100000):
    data = {}
    ids=[]
    for i in range(num_columns):
        column_name = f"column_{i+1}"
        data[column_name] = [random.randint(min_value, max_value) for _ in range(num_rows)]
        ids.append(i)
    return data


def generate_pyarrow_update_data(num_columns=100, num_rows=1000, min_value=0, max_value=100):
    data = {}
    ids=[]
    for i in range(num_columns):
        column_name = f"column_{i+1}"
        data[column_name] = [random.randint(min_value, max_value) for _ in range(num_rows)]
        ids.append(i)
    data['id'] = ids
    return data


def benchmark_read_write_update(num_rows):
    
    update_data=generate_update_data(n_rows=num_rows)
    update_data=db._construct_table(update_data)
    start_time = time.time()
    db.update(update_data,
              )
    update_time = time.time() - start_time
    
    del update_data
    return update_time




if __name__ == '__main__':
    save_dir=os.path.join(config.data_dir, 'benchmarks', 'parquetdb')
    os.makedirs(save_dir, exist_ok=True)

    benchmark_dir=os.path.join(save_dir, 'BenchmarkDB')
    dataset_dir=os.path.join(benchmark_dir, 'parquetdb')

    if os.path.exists(dataset_dir):
        shutil.rmtree(dataset_dir)


    db = ParquetDB(dataset_name='parquetdb', dir=benchmark_dir)
    if db.dataset_exists():
        db.drop_dataset()  # Assuming there's a method to clear the database
    data=generate_data(n_rows=1000000)
    
    start_time=time.time()
    db.create(data)
    insert_time = time.time() - start_time
    del data
    
    start_time=time.time()
    data = db.read()
    read_time = time.time() - start_time
    del data


    row_counts = [1, 10, 100, 1000, 10000, 100000, 1000000]
    insertion_times = []
    reading_times = []
    update_times=[]
    
    
    db.normalize(
        
    )


    results={
        'create_times':[],
        'read_times':[],
        'update_times':[],
        'n_rows':[]
    }
    for num_rows in row_counts:
        print(f'Benchmarking {num_rows} rows...')
        
        update_time = benchmark_read_write_update(num_rows)
        insertion_times.append(insert_time)
        reading_times.append(read_time)
        update_times.append(update_time)
        
        print(f'Insertion Time for {num_rows} rows: {insert_time:.4f} seconds')
        print(f'Reading Time for {num_rows} rows: {read_time:.4f} seconds')
        print(f"Updating Time for {num_rows} rows: {update_time:.4f} seconds")
        print('---')
        
    results['create_times']=insertion_times
    results['read_times']=reading_times
    results['update_times']=update_times
    results['n_rows']=row_counts

    df=pd.DataFrame(results)
    df.to_csv(os.path.join(save_dir, 'parquetdb_update_into_constant_rows_table_input_benchmark.csv'), index=False)


import os
import random
import time
from pymongo import MongoClient
import pandas as pd
from parquetdb import config

def generate_needle_in_haystack_data(n_rows=100, n_columns=100, value=-1):
    random_index = random.randint(0, n_rows-1)
    data=[]

    for j in range(n_rows):
        if j==random_index:
            data.append({f'col_{i}':value for i in range(n_columns)})
        else:
            data.append({f'col_{i}':random.randint(0, 100000) for i in range(n_columns)})
    return data

def remove_db(client, db_name):
    client.drop_database(db_name)

def benchmark_needle_in_haystack_serial(num_rows, client, db_name='benchmark', use_index=True):
    # Remove existing database to start fresh
    remove_db(client, db_name)
    
    # Generate data
    data = generate_needle_in_haystack_data(num_rows)
    
    # Start timing insertion
    start_time = time.time()
    
    # Insert data
    db = client[db_name]
    collection = db['test_collection']
    collection.insert_many(data)
    
    insert_time = time.time() - start_time
    data=None
    
    # Create an index on the 'id' field (index creation time not included in update time)
    if use_index:
        collection.create_index('col_0')
    #####################################################################################
    
    # Start timing reading
    start_time = time.time()
    
    # Execute read query
    # cursor = collection.find({})
    cursor = collection.find({"col_0": -1}, {'col_0': -1})
    results=list(cursor)  # Materialize the cursor
    assert results[0]['col_0'] == -1
    read_time = time.time() - start_time
    
    return insert_time, read_time



def benchmark_needle_in_haystack_with_index():
    save_dir = os.path.join(config.data_dir, 'benchmarks', 'mongodb')
    os.makedirs(save_dir, exist_ok=True)
    use_index=True
    
    # Connect to MongoDB
    client = MongoClient('mongodb://localhost:27017/')
    
    row_counts = [1, 10, 100, 1000, 10000, 100000, 1000000]
    insertion_times = []
    reading_times = []

    results = {
        'create_times': [],
        'read_times': [],
        'n_rows': []
    }

    for num_rows in row_counts:
        print(f'Benchmarking {num_rows} rows...')
        
        insert_time, read_time = benchmark_needle_in_haystack_serial(num_rows, client, use_index=use_index)
        insertion_times.append(insert_time)
        reading_times.append(read_time)
        print(f'Insertion Time for {num_rows} rows: {insert_time:.4f} seconds')
        print(f'Reading Time for {num_rows} rows: {read_time:.4f} seconds')
        print('---')
        
    results['create_times'] = insertion_times
    results['read_times'] = reading_times
    results['n_rows'] = row_counts

    df = pd.DataFrame(results)
    df.to_csv(os.path.join(save_dir, 'mongodb_needle_in_haystack_with_index_benchmark.csv'), index=False)

    # Close the MongoDB connection
    client.close()


def benchmark_needle_in_haystack_without_index():
    save_dir = os.path.join(config.data_dir, 'benchmarks', 'mongodb')
    os.makedirs(save_dir, exist_ok=True)
    use_index=False
    
    # Connect to MongoDB
    client = MongoClient('mongodb://localhost:27017/')
    
    row_counts = [1, 10, 100, 1000, 10000, 100000, 1000000]
    insertion_times = []
    reading_times = []

    results = {
        'create_times': [],
        'read_times': [],
        'n_rows': []
    }

    for num_rows in row_counts:
        print(f'Benchmarking {num_rows} rows...')
        
        insert_time, read_time = benchmark_needle_in_haystack_serial(num_rows, client, use_index=use_index)
        insertion_times.append(insert_time)
        reading_times.append(read_time)
        print(f'Insertion Time for {num_rows} rows: {insert_time:.4f} seconds')
        print(f'Reading Time for {num_rows} rows: {read_time:.4f} seconds')
        print('---')
        
    results['create_times'] = insertion_times
    results['read_times'] = reading_times
    results['n_rows'] = row_counts

    df = pd.DataFrame(results)
    df.to_csv(os.path.join(save_dir, 'mongodb_needle_in_haystack_without_index_benchmark.csv'), index=False)

    # Close the MongoDB connection
    client.close()
    

if __name__ == '__main__':
    
    
    benchmark_needle_in_haystack_with_index()
    benchmark_needle_in_haystack_without_index()


import os
import shutil
import sys
import time
import random
import string
import itertools

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1.inset_locator import inset_axes

import pyarrow as pa
import pyarrow.compute as pc
import pyarrow.dataset as ds
import pyarrow.parquet as pq
import pyarrow.fs as fs

from parquetdb import ParquetDB, config

import matplotlib.ticker as ticker

plt.rcParams['axes.labelsize'] = 18
plt.rcParams['axes.titlesize'] = 18
plt.rcParams['xtick.labelsize'] = 14
plt.rcParams['ytick.labelsize'] = 14


db_names=['sqlite','mongodb','parquetdb']
benchmark_dir=os.path.join(config.data_dir, 'benchmarks')
benchmark_dirs=[os.path.join(benchmark_dir, db_name) for db_name in db_names]

benchmark_type='needle_in_haystack'


benchmark_files_dict={
    'sqlite':['sqlite_needle_in_haystack_with_index_benchmark.csv','sqlite_needle_in_haystack_without_index_benchmark.csv'],
    'mongodb':['mongodb_needle_in_haystack_with_index_benchmark.csv','mongodb_needle_in_haystack_without_index_benchmark.csv'],
    'parquetdb':['parquetdb_needle_in_haystack_benchmark.csv']
}

line_styles={
    'sqlite': {'with_index': 'solid', 'without_index': 'dashed'},
    'mongodb': {'with_index': 'solid', 'without_index': 'dashed'},
    'parquetdb': 'solid'
}
            
def color_diff_log_inset_plot(savefig=None):
    # Initialize a figure and axes
    fig, ax1 = plt.subplots(figsize=(10, 6))


    # Define a color map for different databases
    # colors = {
    #     'sqlite': "#3B9AB2",
    #     'mongodb': "#EBCC2A",
    #     'parquetdb': "#F21A00"
    # }
    colors = {
        'sqlite': "#FF0000",
        'mongodb': "#F2AD00",
        'parquetdb': "#5BBCD6"
    }
    
    

    # Loop through each database benchmark data and plot
    for db_name, benchmark_files in benchmark_files_dict.items():
        for benchmark_file in benchmark_files:
            basename = os.path.basename(benchmark_file)
            csv_filename = os.path.join(benchmark_dir,db_name, benchmark_file)
            
            df = pd.read_csv(csv_filename)

            if 'with_index' in benchmark_file:
                line_style=line_styles[db_name]['with_index']
                label=f'{db_name} with index'
            elif 'without_index' in benchmark_file:
                line_style=line_styles[db_name]['without_index']
                label=f'{db_name} without index'
            else:
                line_style=line_styles[db_name]
                label=f'{db_name}'
            
            # Plot create times on the primary y-axis
            ax1.plot(df['n_rows'], df['read_times'], label=label, color=colors[db_name], linestyle=line_style)

    # Create twin axis for the secondary y-axis
   

    # Set labels and title
    ax1.set_xlabel('Number of Rows')
    ax1.set_ylabel('Read Times (s)')
    # ax2.set_ylabel('Read Times (s)')
    
    # Ensure both axes have a common scale by linking their limits
    # ax1.set_xscale('log')
    # ax1.set_yscale('log')

    # Color the spines and tick labels
    # ax1.spines['left'].set_color('blue')
    # ax1.tick_params(axis='y', colors='blue')

    # Set the same linestyle and make the spine thicker for visibility
    ax1.spines['left'].set_linestyle('solid')
    ax1.spines['left'].set_linewidth(2.5)  # Increase the line width for visibility


    # Hide the right spine on ax1 and left spine on ax2 to prevent overlap
    # ax1.spines['right'].set_visible(False)


    ax1.tick_params(axis='both', which='major', length=10, width=2, direction='out')

    ax1.grid(True)
    
    scale=36
    ax_inset = inset_axes(ax1, width=f"{scale}%", height=f"{scale}%", loc="upper left", bbox_to_anchor=(0.05, -0.03,1,1), bbox_transform=ax1.transAxes, borderpad=2)

    # Add an inset with log scale
    # ax_inset = inset_axes(ax1, width="30%", height="30%", loc="upper left", bbox_to_anchor=(0,0,1,1))
    # ax_inset = inset_axes(ax1, width="30%", height="30%", bbox_to_anchor=(0,0,1,1))
    ax_inset.grid(True)
    # Plot the same data with log scales
    for db_name, benchmark_files in benchmark_files_dict.items():
        for benchmark_file in benchmark_files:
            basename = os.path.basename(benchmark_file)
            csv_filename = os.path.join(benchmark_dir,db_name, benchmark_file)
            
            df = pd.read_csv(csv_filename)

            if 'with_index' in benchmark_file:
                line_style=line_styles[db_name]['with_index']
                label=f'{db_name} with index'
            elif 'without_index' in benchmark_file:
                line_style=line_styles[db_name]['without_index']
                label=f'{db_name} without index'
            else:
                line_style=line_styles[db_name]
                label=f'{db_name}'
            
            ax_inset.plot(df['n_rows'], df['read_times'], label=label, color=colors[db_name], linestyle=line_style,  linewidth=2)


    # Set log scale for both axes in the inset
    ax_inset.set_xscale('log')
    ax_inset.set_yscale('log')

    # ax_inset.xaxis.set_major_locator(LogLocator(base=10.0, subs=None, numticks=6))
    # ax_inset.yaxis.set_major_locator(LogLocator(base=10.0, subs=None, numticks=10))
    nticks = 9
    maj_loc = ticker.LogLocator(numticks=nticks)
    min_loc = ticker.LogLocator(subs='all', numticks=nticks)
    ax_inset.xaxis.set_major_locator(maj_loc)
    ax_inset.xaxis.set_minor_locator(min_loc)

    # Set labels for inset plot
    ax_inset.set_xlabel('Number of Rows (log)', fontsize=8)
    ax_inset.set_ylabel('Read Time (log)', fontsize=8, labelpad=-2)  # Added labelpad to move label right
    # ax_inset2.set_ylabel('Read Time (log)', fontsize=8)
    # Set x-axis ticks to show all orders of magnitude
    # ax_inset.set_xticks([10**1, 10**2, 10**3, 10**4, 10**5])
    # ax_inset.set_xticklabels(['$10^1$', '$10^2$', '$10^3$', '$10^4$', '$10^5$', '$10^5$'])
    
    # Set the same linestyle and make the spine thicker for visibility
    ax_inset.spines['left'].set_linestyle('solid')
    ax_inset.spines['left'].set_linewidth(2.5)  # Increase the line width for visibility


    # Hide the right spine on ax1 and left spine on ax2 to prevent overlap
    # ax_inset.spines['right'].set_visible(False)

    ax_inset.tick_params(axis='both', which='major', length=6, width=1.5, direction='out')
    ax_inset.tick_params(axis='x', which='minor', length=3, width=1, direction='out')
    ax_inset.tick_params(axis='y', which='minor', length=3, width=1, direction='out')

    lines_1, labels_1 = ax1.get_legend_handles_labels()

    ax1.legend(lines_1, labels_1 , loc='upper center',  bbox_to_anchor=(0.12, 0,1,1))
    
    ax1.set_title('Needle in Hay Stack Query Benchmark \n for SQLite, MongoDB, and ParquetDB for 100 integer columns')
    plt.tight_layout()
    
    if savefig:
        plt.savefig(savefig)
        return None
    plt.show()
    
    

    
if __name__ == '__main__':
    from parquetdb.utils import matplotlib_utils
    matplotlib_utils.set_palette('Zissou1')

    # color_diff_log_inset_plot()
    color_diff_log_inset_plot(savefig=os.path.join(benchmark_dir, 'benchmark_needle_in_haystack_times.pdf'))
    color_diff_log_inset_plot(savefig=os.path.join(benchmark_dir, 'benchmark_needle_in_haystack_times.png'))
    


from setuptools import setup
from setuptools_scm import ScmVersion

def version_for_project(version: ScmVersion) -> str:
   return str(version.tag)
   
setup(use_scm_version={'version_scheme':version_for_project})



import os
import random
import time
from pymongo import MongoClient, UpdateOne, UpdateMany
import pandas as pd
from parquetdb import config

def generate_data(num_rows, n_columns=100):
    data = []
    for i in range(num_rows):
        doc = {'id': i}  # Unique identifier for each document
        doc.update({f'col{j}': random.randint(0, 1000000) for j in range(n_columns)})
        data.append(doc)
    return data

def remove_db(client, db_name):
    client.drop_database(db_name)

def benchmark_read_write_update(num_rows, client, db_name='benchmark'):
    ################################################################################################
    db = client[db_name]
    collection = db['test_collection']
    
    ################################################################################################
    # Prepare update data with the same 'id's
    update_data = generate_data(num_rows)
    
    # Prepare bulk update operations
    update_operations = []
    for doc in update_data:
        doc_id = doc['id']
        update_fields = {k: v for k, v in doc.items() if k != 'id'}
        update_operations.append(
            UpdateOne({'id': doc_id}, {'$set': update_fields})
        )
    del update_data
    # Measure update time
    start_time = time.time()
    
    # Perform bulk update operations
    if update_operations:
        collection.bulk_write(update_operations)
    
    update_time = time.time() - start_time

    return update_time

if __name__ == '__main__':
    save_dir = os.path.join(config.data_dir, 'benchmarks', 'mongodb')
    os.makedirs(save_dir, exist_ok=True)
    
    # Connect to MongoDB
    client = MongoClient('mongodb://localhost:27017/')
    
    db_name='benchmark'
    
    
    # Remove existing database to start fresh
    remove_db(client, db_name)
    
    # Generate initial data
    data = generate_data(num_rows=1000000)
    
    # Start timing insertion
    start_time = time.time()
    
    # Insert data
    db = client[db_name]
    collection = db['test_collection']
    collection.insert_many(data)
    
    insert_time = time.time() - start_time
    data = None  # Release memory
    
    # Create an index on the 'id' field (index creation time not included in update time)
    collection.create_index('id', unique=True)
    ################################################################################################
    # Start timing reading
    start_time = time.time()
    
    # Execute read query
    cursor = collection.find({})
    list(cursor)  # Materialize the cursor
    read_time = time.time() - start_time
    ################################################################################################
    
    
    row_counts = [1, 10, 100, 1000, 10000, 100000, 1000000]
    insertion_times = []
    reading_times = []
    update_times = []

    results = {
        'create_times': [],
        'read_times': [],
        'update_times': [],
        'n_rows': []
    }

    for num_rows in row_counts:
        print(f'Benchmarking {num_rows} rows...')
        
        update_time = benchmark_read_write_update(num_rows, client)
        insertion_times.append(insert_time)
        reading_times.append(read_time)
        update_times.append(update_time)
        print(f'Insertion Time for {num_rows} rows: {insert_time:.4f} seconds')
        print(f'Reading Time for {num_rows} rows: {read_time:.4f} seconds')
        print(f'Update Time for {num_rows} rows: {update_time:.4f} seconds')
        print('---')
        
    results['create_times'] = insertion_times
    results['read_times'] = reading_times
    results['update_times'] = update_times
    results['n_rows'] = row_counts

    df = pd.DataFrame(results)
    df.to_csv(os.path.join(save_dir, 'mongodb_update_into_constant_rows_benchmark.csv'), index=False)

    # Close the MongoDB connection
    client.close()